from itertools import product

import jax.numpy as jnp
import numpy as np
import pytest
from numpy.linalg import norm
from numpy.testing import assert_, assert_allclose, assert_equal, suppress_warnings
from pytest import raises as assert_raises
from scipy.sparse import lil_matrix
from scipy.sparse.linalg import aslinearoperator

from nlsq.core.least_squares import LeastSquares

lsqs = LeastSquares()
least_squares = lsqs.least_squares
from scipy.optimize._lsq.common import EPS, make_strictly_feasible
from scipy.optimize._lsq.least_squares import IMPLEMENTED_LOSSES


def fun_trivial(x, a=0):
    return (x[0] - a) ** 2 + 5.0


def jac_trivial(x, a=0.0):
    return 2 * (x[0] - a)


def fun_2d_trivial(x):
    return jnp.array([x[0], x[1]])


def jac_2d_trivial(x):
    return jnp.identity(2)


def fun_rosenbrock(x):
    return jnp.array([10 * (x[1] - x[0] ** 2), (1 - x[0])])


def jac_rosenbrock(x):
    return jnp.array([[-20 * x[0], 10], [-1, 0]])


def jac_rosenbrock_bad_dim(x):
    return jnp.array([[-20 * x[0], 10], [-1, 0], [0.0, 0.0]])


def fun_rosenbrock_cropped(x):
    return fun_rosenbrock(x)[0]


def jac_rosenbrock_cropped(x):
    return jac_rosenbrock(x)[:1, :]


# When x is 1-D array, return is 2-D array.
def fun_wrong_dimensions(x):
    x = jnp.array(x)
    return jnp.array([x, x**2, x**3])


# need transpose here
def jac_wrong_dimensions(x, a=0.0):
    return np.atleast_3d(jac_trivial(x, a=a))


# def fun_bvp(x):
#     n = int(np.sqrt(x.shape[0]))
#     u = np.zeros((n + 2, n + 2))
#     x = x.reshape((n, n))
#     u[1:-1, 1:-1] = x
#     y = u[:-2, 1:-1] + u[2:, 1:-1] + u[1:-1, :-2] + u[1:-1, 2:] - 4 * x + x**3
#     return y.ravel()


def fun_bvp(x):
    x = jnp.array(x)
    n = int(len(x) ** 0.5)
    u = jnp.zeros((n + 2, n + 2))
    x = x.reshape((n, n))
    u = u.at[1:-1, 1:-1].set(x)
    y = u[:-2, 1:-1] + u[2:, 1:-1] + u[1:-1, :-2] + u[1:-1, 2:] - 4 * x + x**3
    return y.ravel()


class BroydenTridiagonal:
    def __init__(self, n=100, mode="sparse"):
        np.random.seed(0)

        self.n = n

        self.x0 = -np.ones(n)
        self.lb = np.linspace(-2, -1.5, n)
        self.ub = np.linspace(-0.8, 0.0, n)

        self.lb += 0.1 * np.random.randn(n)
        self.ub += 0.1 * np.random.randn(n)

        self.x0 += 0.1 * np.random.randn(n)
        self.x0 = make_strictly_feasible(self.x0, self.lb, self.ub)

        if mode == "sparse":
            self.sparsity = lil_matrix((n, n), dtype=int)
            i = np.arange(n)
            self.sparsity[i, i] = 1
            i = np.arange(1, n)
            self.sparsity[i, i - 1] = 1
            i = np.arange(n - 1)
            self.sparsity[i, i + 1] = 1

            self.jac = self._jac
        elif mode == "operator":
            self.jac = lambda x: aslinearoperator(self._jac(x))
        elif mode == "dense":
            self.sparsity = None
            self.jac = lambda x: self._jac(x).toarray()
        else:
            assert_(False)

    def fun(self, x):
        f = (3 - x) * x + 1
        f[1:] -= x[:-1]
        f[:-1] -= 2 * x[1:]
        return f

    def _jac(self, x):
        J = lil_matrix((self.n, self.n))
        i = np.arange(self.n)
        J[i, i] = 3 - 2 * x
        i = np.arange(1, self.n)
        J[i, i - 1] = -1
        i = np.arange(self.n - 1)
        J[i, i + 1] = -2
        return J


class ExponentialFittingProblem:
    """Provide data and function for exponential fitting in the form
    y = a + exp(b * x) + noise."""

    def __init__(
        self, a, b, noise, n_outliers=1, x_range=(-1, 1), n_points=11, random_seed=None
    ):
        np.random.seed(random_seed)
        self.m = n_points
        self.n = 2

        self.p0 = np.zeros(2)
        self.x = np.linspace(x_range[0], x_range[1], n_points)

        self.y = a + np.exp(b * self.x)
        self.y += noise * np.random.randn(self.m)

        outliers = np.random.randint(0, self.m, n_outliers)
        self.y[outliers] += 50 * noise * np.random.rand(n_outliers)

        self.p_opt = np.array([a, b])

    def fun(self, p):
        return p[0] + jnp.exp(p[1] * self.x) - self.y

    def jac(self, p):
        J = np.empty((self.m, self.n))
        J[:, 0] = 1
        J[:, 1] = self.x * np.exp(p[1] * self.x)
        return J


def cubic_soft_l1(z):
    rho = np.empty((3, z.size))

    t = 1 + z
    rho[0] = 3 * (t ** (1 / 3) - 1)
    rho[1] = t ** (-2 / 3)
    rho[2] = -2 / 3 * t ** (-5 / 3)

    return rho


LOSSES = [*list(IMPLEMENTED_LOSSES.keys()), cubic_soft_l1]


class BaseMixin:
    def test_basic(self):
        # Test that the basic calling sequence works.
        res = least_squares(fun_trivial, 2.0, method=self.method)
        assert_allclose(res.x, 0, atol=1e-4)
        assert_allclose(res.fun, fun_trivial(res.x))

    def test_args_kwargs(self):
        # Test that args and kwargs are passed correctly to the functions.
        a = 3.0
        for jac in [None, jac_trivial]:
            with suppress_warnings() as sup:
                sup.filter(
                    UserWarning,
                    "jac='(3-point|cs)' works equivalently to '2-point' for method='lm'",
                )
                res = least_squares(
                    fun_trivial, 2.0, jac, args=(a,), method=self.method
                )
                res1 = least_squares(
                    fun_trivial, 2.0, jac, kwargs={"a": a}, method=self.method
                )

            assert_allclose(res.x, a, rtol=1e-4)
            assert_allclose(res1.x, a, rtol=1e-4)

            assert_raises(
                TypeError,
                least_squares,
                fun_trivial,
                2.0,
                args=(
                    3,
                    4,
                ),
                method=self.method,
            )
            assert_raises(
                TypeError,
                least_squares,
                fun_trivial,
                2.0,
                kwargs={"kaboom": 3},
                method=self.method,
            )

    def test_jac_options(self):
        for jac in [None, jac_trivial]:
            with suppress_warnings() as sup:
                sup.filter(
                    UserWarning,
                    "jac='(3-point|cs)' works equivalently to '2-point' for method='lm'",
                )
                res = least_squares(fun_trivial, 2.0, jac, method=self.method)
            assert_allclose(res.x, 0, atol=1e-4)

        assert_raises(
            ValueError, least_squares, fun_trivial, 2.0, jac="oops", method=self.method
        )

    def test_nfev_options(self):
        for max_nfev in [None, 20]:
            res = least_squares(fun_trivial, 2.0, max_nfev=max_nfev, method=self.method)
            assert_allclose(res.x, 0, atol=1e-4)

    def test_x_scale_options(self):
        for x_scale in [1.0, np.array([0.5]), "jac"]:
            res = least_squares(fun_trivial, 2.0, x_scale=x_scale)
            assert_allclose(res.x, 0)
        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            x_scale="auto",
            method=self.method,
        )
        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            x_scale=-1.0,
            method=self.method,
        )
        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            x_scale=None,
            method=self.method,
        )
        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            x_scale=1.0 + 2.0j,
            method=self.method,
        )

    def test_diff_step(self):
        # res1 and res2 should be equivalent.
        # res2 and res3 should be different.
        res1 = least_squares(fun_trivial, 2.0, diff_step=1e-1, method=self.method)
        res2 = least_squares(fun_trivial, 2.0, diff_step=-1e-1, method=self.method)
        res3 = least_squares(fun_trivial, 2.0, diff_step=None, method=self.method)
        assert_allclose(res1.x, 0, atol=1e-4)
        assert_allclose(res2.x, 0, atol=1e-4)
        assert_allclose(res3.x, 0, atol=1e-4)
        assert_equal(res1.x, res2.x)
        assert_equal(res1.nfev, res2.nfev)

    def test_incorrect_options_usage(self):
        assert_raises(
            TypeError,
            least_squares,
            fun_trivial,
            2.0,
            method=self.method,
            options={"no_such_option": 100},
        )
        assert_raises(
            TypeError,
            least_squares,
            fun_trivial,
            2.0,
            method=self.method,
            options={"max_nfev": 100},
        )

    def test_full_result(self):
        # MINPACK doesn't work very well with factor=100 on this problem,
        # thus using low 'atol'.
        res = least_squares(fun_trivial, 2.0, method=self.method)
        assert_allclose(res.x, 0, atol=1e-4)
        assert_allclose(res.cost, 12.5)
        assert_allclose(res.fun, 5)
        assert_allclose(res.jac, 0, atol=1e-4)
        assert_allclose(res.grad, 0, atol=1e-2)
        assert_allclose(res.optimality, 0, atol=1e-2)
        assert_equal(res.active_mask, np.array([0.0]))
        if self.method == "lm":
            assert_(res.nfev < 30)
            assert_(res.njev is None)
        else:
            assert_(res.nfev < 10)
            assert_(res.njev < 10)
        assert_(res.status > 0)
        assert_(res.success)

    def test_full_result_single_fev(self):
        # MINPACK checks the number of nfev after the iteration,
        # so it's hard to tell what he is going to compute.
        if self.method == "lm":
            return

        res = least_squares(fun_trivial, 2.0, method=self.method, max_nfev=1)
        assert_equal(res.x, np.array([2]))
        assert_equal(res.cost, 40.5)
        assert_equal(res.fun, np.array([9]))
        assert_equal(res.jac, np.array([[4]]))
        assert_equal(res.grad, np.array([36]))
        assert_equal(res.optimality, 36)
        assert_equal(res.active_mask, np.array([0]))
        assert_equal(res.nfev, 1)
        assert_equal(res.njev, 1)
        assert_equal(res.status, 0)
        assert_equal(res.success, 0)

    def test_rosenbrock(self):
        x0 = [-2, 1]
        x_opt = [1, 1]
        for jac, x_scale, tr_solver in product(
            [None, jac_rosenbrock],
            [1.0, np.array([1.0, 0.2]), "jac"],
            ["exact", "lsmr"],
        ):
            with suppress_warnings() as sup:
                sup.filter(
                    UserWarning,
                    "jac='(3-point|cs)' works equivalently to '2-point' for method='lm'",
                )
                res = least_squares(
                    fun_rosenbrock,
                    x0,
                    jac,
                    x_scale=x_scale,
                    tr_solver=tr_solver,
                    method=self.method,
                )
            assert_allclose(res.x, x_opt)

    def test_rosenbrock_cropped(self):
        x0 = [-2, 1]
        if self.method == "lm":
            assert_raises(
                ValueError, least_squares, fun_rosenbrock_cropped, x0, method="lm"
            )
        else:
            for jac, x_scale, tr_solver in product(
                [None, jac_rosenbrock_cropped],
                [1.0, np.array([1.0, 0.2]), "jac"],
                ["exact", "lsmr"],
            ):
                res = least_squares(
                    fun_rosenbrock_cropped,
                    x0,
                    jac,
                    x_scale=x_scale,
                    tr_solver=tr_solver,
                    method=self.method,
                )
                assert_allclose(res.cost, 0, atol=1e-14)

    def test_fun_wrong_dimensions(self):
        assert_raises(
            ValueError, least_squares, fun_wrong_dimensions, 2.0, method=self.method
        )

    def test_jac_wrong_dimensions(self):
        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            jac_wrong_dimensions,
            method=self.method,
        )

    # def test_fun_and_jac_inconsistent_dimensions(self):
    #     x0 = [1, 2]
    #     assert_raises(ValueError, least_squares, fun_rosenbrock, x0,
    #                   jac_rosenbrock_bad_dim, method=self.method)

    def test_x0_multidimensional(self):
        x0 = np.ones(4).reshape(2, 2)
        assert_raises(ValueError, least_squares, fun_trivial, x0, method=self.method)

    def test_x0_complex_scalar(self):
        x0 = 2.0 + 0.0 * 1j
        assert_raises(ValueError, least_squares, fun_trivial, x0, method=self.method)

    def test_x0_complex_array(self):
        x0 = [1.0, 2.0 + 0.0 * 1j]
        assert_raises(ValueError, least_squares, fun_trivial, x0, method=self.method)

    def test_bvp(self):
        # This test was introduced with fix #5556. It turned out that
        # dogbox solver had a bug with trust-region radius update, which
        # could block its progress and create an infinite loop. And this
        # discrete boundary value problem is the one which triggers it.
        n = 10
        x0 = np.ones(n**2)
        if self.method == "lm":
            max_nfev = 5000  # To account for Jacobian estimation.
        else:
            max_nfev = 100
        res = least_squares(
            fun_bvp, x0, ftol=1e-2, method=self.method, max_nfev=max_nfev
        )

        assert_(res.nfev < max_nfev)
        assert_(res.cost < 0.5)

    def test_error_raised_when_all_tolerances_below_eps(self):
        # Test that all 0 tolerances are not allowed.
        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            method=self.method,
            ftol=None,
            xtol=None,
            gtol=None,
        )

    def test_convergence_with_only_one_tolerance_enabled(self):
        if self.method == "lm":
            return  # should not do test
        x0 = [-2, 1]
        x_opt = [1, 1]
        for ftol, xtol, gtol in [
            (1e-8, None, None),
            (None, 1e-8, None),
            (None, None, 1e-8),
        ]:
            res = least_squares(
                fun_rosenbrock,
                x0,
                jac=jac_rosenbrock,
                ftol=ftol,
                gtol=gtol,
                xtol=xtol,
                method=self.method,
            )
            assert_allclose(res.x, x_opt)


class BoundsMixin:
    def test_inconsistent(self):
        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            bounds=(10.0, 0.0),
            method=self.method,
        )

    def test_infeasible(self):
        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            bounds=(3.0, 4),
            method=self.method,
        )

    def test_wrong_number(self):
        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            bounds=(1.0, 2, 3),
            method=self.method,
        )

    def test_inconsistent_shape(self):
        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            bounds=(1.0, [2.0, 3.0]),
            method=self.method,
        )
        # 1-D array won't be broadcasted
        assert_raises(
            ValueError,
            least_squares,
            fun_rosenbrock,
            [1.0, 2.0],
            bounds=([0.0], [3.0, 4.0]),
            method=self.method,
        )

    def test_in_bounds(self):
        for jac in [None, jac_trivial]:
            res = least_squares(
                fun_trivial, 2.0, jac=jac, bounds=(-1.0, 3.0), method=self.method
            )
            assert_allclose(res.x, 0.0, atol=1e-4)
            assert_equal(res.active_mask, [0])
            assert_(-1 <= res.x <= 3)
            res = least_squares(
                fun_trivial, 2.0, jac=jac, bounds=(0.5, 3.0), method=self.method
            )
            assert_allclose(res.x, 0.5, atol=1e-4)
            assert_equal(res.active_mask, [-1])
            assert_(0.5 <= res.x <= 3)

    def test_bounds_shape(self):
        for jac in [None, jac_2d_trivial]:
            x0 = [1.0, 1.0]
            res = least_squares(fun_2d_trivial, x0, jac=jac)
            assert_allclose(res.x, [0.0, 0.0])
            res = least_squares(
                fun_2d_trivial,
                x0,
                jac=jac,
                bounds=(0.5, [2.0, 2.0]),
                method=self.method,
            )
            assert_allclose(res.x, [0.5, 0.5])
            res = least_squares(
                fun_2d_trivial,
                x0,
                jac=jac,
                bounds=([0.3, 0.2], 3.0),
                method=self.method,
            )
            assert_allclose(res.x, [0.3, 0.2])
            res = least_squares(
                fun_2d_trivial,
                x0,
                jac=jac,
                bounds=([-1, 0.5], [1.0, 3.0]),
                method=self.method,
            )
            assert_allclose(res.x, [0.0, 0.5], atol=1e-5)

    def test_rosenbrock_bounds(self):
        x0_1 = np.array([-2.0, 1.0])
        x0_2 = np.array([2.0, 2.0])
        x0_3 = np.array([-2.0, 2.0])
        x0_4 = np.array([0.0, 2.0])
        x0_5 = np.array([-1.2, 1.0])
        problems = [
            (x0_1, ([-np.inf, -1.5], np.inf)),
            (x0_2, ([-np.inf, 1.5], np.inf)),
            (x0_3, ([-np.inf, 1.5], np.inf)),
            (x0_4, ([-np.inf, 1.5], [1.0, np.inf])),
            (x0_2, ([1.0, 1.5], [3.0, 3.0])),
            (x0_5, ([-50.0, 0.0], [0.5, 100])),
        ]
        for x0, bounds in problems:
            for jac, x_scale, tr_solver in product(
                [None, jac_rosenbrock], [1.0, [1.0, 0.5], "jac"], ["exact", "lsmr"]
            ):
                res = least_squares(
                    fun_rosenbrock,
                    x0,
                    jac,
                    bounds,
                    x_scale=x_scale,
                    tr_solver=tr_solver,
                    method=self.method,
                )
                assert_allclose(res.optimality, 0.0, atol=1e-5)


class SparseMixin:
    def test_exact_tr_solver(self):
        p = BroydenTridiagonal()
        assert_raises(
            ValueError,
            least_squares,
            p.fun,
            p.x0,
            p.jac,
            tr_solver="exact",
            method=self.method,
        )
        # assert_raises(ValueError, least_squares, p.fun, p.x0,
        #               tr_solver='exact', jac_sparsity=p.sparsity,
        #               method=self.method)

    # def test_equivalence(self):
    #     sparse = BroydenTridiagonal(mode='sparse')
    #     dense = BroydenTridiagonal(mode='dense')
    #     res_sparse = least_squares(
    #         sparse.fun, sparse.x0, jac=sparse.jac,
    #         method=self.method)
    #     res_dense = least_squares(
    #         dense.fun, dense.x0, jac=sparse.jac,
    #         method=self.method)
    #     assert_equal(res_sparse.nfev, res_dense.nfev)
    #     assert_allclose(res_sparse.x, res_dense.x, atol=1e-20)
    #     assert_allclose(res_sparse.cost, 0, atol=1e-20)
    #     assert_allclose(res_dense.cost, 0, atol=1e-20)

    def test_tr_options(self):
        p = BroydenTridiagonal()
        res = least_squares(
            p.fun, p.x0, p.jac, method=self.method, tr_options={"btol": 1e-10}
        )
        assert_allclose(res.cost, 0, atol=1e-20)

    def test_wrong_parameters(self):
        p = BroydenTridiagonal()
        assert_raises(
            ValueError,
            least_squares,
            p.fun,
            p.x0,
            p.jac,
            tr_solver="best",
            method=self.method,
        )
        assert_raises(
            TypeError,
            least_squares,
            p.fun,
            p.x0,
            p.jac,
            tr_solver="lsmr",
            tr_options={"tol": 1e-10},
        )

    def test_solver_selection(self):
        # sparse = BroydenTridiagonal(mode='sparse')
        dense = BroydenTridiagonal(mode="dense")
        # res_sparse = least_squares(sparse.fun, sparse.x0, jac=sparse.jac,
        #                             method=self.method)
        res_dense = least_squares(
            dense.fun, dense.x0, jac=dense.jac, method=self.method
        )
        # assert_allclose(res_sparse.cost, 0, atol=1e-20)
        assert_allclose(res_dense.cost, 0, atol=1e-20)
        # assert_(issparse(res_sparse.jac))
        assert_(isinstance(res_dense.jac, np.ndarray))

    def test_numerical_jac(self):
        p = BroydenTridiagonal()
        for jac in [None]:
            res_dense = least_squares(p.fun, p.x0, jac, method=self.method)
            # res_sparse = least_squares(
            #     p.fun, p.x0, jac,method=self.method,
            #     jac_sparsity=p.sparsity)
            # assert_equal(res_dense.nfev, res_sparse.nfev)
            # assert_allclose(res_dense.x, res_sparse.x, atol=1e-20)
            assert_allclose(res_dense.cost, 0, atol=1e-20)
            # assert_allclose(res_sparse.cost, 0, atol=1e-20)

    def test_with_bounds(self):
        p = BroydenTridiagonal()
        for jac, jac_sparsity in product([p.jac, None], [None]):
            res_1 = least_squares(
                p.fun,
                p.x0,
                jac,
                bounds=(p.lb, np.inf),
                method=self.method,
                jac_sparsity=jac_sparsity,
            )
            res_2 = least_squares(
                p.fun,
                p.x0,
                jac,
                bounds=(-np.inf, p.ub),
                method=self.method,
                jac_sparsity=jac_sparsity,
            )
            res_3 = least_squares(
                p.fun,
                p.x0,
                jac,
                bounds=(p.lb, p.ub),
                method=self.method,
                jac_sparsity=jac_sparsity,
            )
            assert_allclose(res_1.optimality, 0, atol=1e-10)
            assert_allclose(res_2.optimality, 0, atol=1e-10)
            assert_allclose(res_3.optimality, 0, atol=1e-10)

    # def test_wrong_jac_sparsity(self):
    #     p = BroydenTridiagonal()
    #     sparsity = p.sparsity[:-1]
    #     assert_raises(ValueError, least_squares, p.fun, p.x0,
    #                   jac_sparsity=sparsity, method=self.method)

    # def test_linear_operator(self):
    #     p = BroydenTridiagonal(mode='operator')
    #     res = least_squares(p.fun, p.x0, p.jac, method=self.method)
    #     assert_allclose(res.cost, 0.0, atol=1e-20)
    #     assert_raises(ValueError, least_squares, p.fun, p.x0, p.jac,
    #                   method=self.method, tr_solver='exact')

    def test_x_scale_jac_scale(self):
        p = BroydenTridiagonal()
        res = least_squares(p.fun, p.x0, p.jac, method=self.method, x_scale="jac")
        assert_allclose(res.cost, 0.0, atol=1e-20)

        # p = BroydenTridiagonal(mode='operator')
        # assert_raises(ValueError, least_squares, p.fun, p.x0, p.jac,
        #               method=self.method, x_scale='jac')


class LossFunctionMixin:
    def test_options(self):
        for loss in LOSSES:
            res = least_squares(fun_trivial, 2.0, loss=loss, method=self.method)
            assert_allclose(res.x, 0, atol=1e-15)

        assert_raises(
            ValueError,
            least_squares,
            fun_trivial,
            2.0,
            loss="hinge",
            method=self.method,
        )

    def test_fun(self):
        # Test that res.fun is actual residuals, and not modified by loss
        # function stuff.
        for loss in LOSSES:
            res = least_squares(fun_trivial, 2.0, loss=loss, method=self.method)
            assert_equal(np.array(res.fun), fun_trivial(res.x))

    def test_grad(self):
        # Test that res.grad is true gradient of loss function at the
        # solution. Use max_nfev = 1, to avoid reaching minimum.
        x = np.array([2.0])  # res.x will be this.

        res = least_squares(
            fun_trivial, x, jac_trivial, loss="linear", max_nfev=1, method=self.method
        )
        assert_equal(res.grad, 2 * x * (x**2 + 5))

        res = least_squares(
            fun_trivial, x, jac_trivial, loss="huber", max_nfev=1, method=self.method
        )

        assert_equal(res.grad, 2 * x)

        res = least_squares(
            fun_trivial, x, jac_trivial, loss="soft_l1", max_nfev=1, method=self.method
        )
        assert_allclose(res.grad, 2 * x * (x**2 + 5) / (1 + (x**2 + 5) ** 2) ** 0.5)

        res = least_squares(
            fun_trivial, x, jac_trivial, loss="cauchy", max_nfev=1, method=self.method
        )
        assert_allclose(res.grad, 2 * x * (x**2 + 5) / (1 + (x**2 + 5) ** 2))

        res = least_squares(
            fun_trivial, x, jac_trivial, loss="arctan", max_nfev=1, method=self.method
        )
        assert_allclose(res.grad, 2 * x * (x**2 + 5) / (1 + (x**2 + 5) ** 4))

        res = least_squares(
            fun_trivial,
            x,
            jac_trivial,
            loss=cubic_soft_l1,
            max_nfev=1,
            method=self.method,
        )
        assert_allclose(res.grad, 2 * x * (x**2 + 5) / (1 + (x**2 + 5) ** 2) ** (2 / 3))

    def test_jac(self):
        # Test that res.jac.T.dot(res.jac) gives Gauss-Newton approximation
        # of Hessian. This approximation is computed by doubly differentiating
        # the cost function and dropping the part containing second derivative
        # of f. For a scalar function it is computed as
        # H = (rho' + 2 * rho'' * f**2) * f'**2, if the expression inside the
        # brackets is less than EPS it is replaced by EPS. Here, we check
        # against the root of H.

        x = 2.0  # res.x will be this.
        f = x**2 + 5  # res.fun will be this.

        res = least_squares(
            fun_trivial, x, jac_trivial, loss="linear", max_nfev=1, method=self.method
        )
        assert_equal(np.array(res.jac)[0, 0], 2 * x)

        # For `huber` loss the Jacobian correction is identically zero
        # in outlier region, in such cases it is modified to be equal EPS**0.5.
        res = least_squares(
            fun_trivial, x, jac_trivial, loss="huber", max_nfev=1, method=self.method
        )
        assert_equal(np.array(res.jac), 2 * x * EPS**0.5)

        # Now, let's apply `loss_scale` to turn the residual into an inlier.
        # The loss function becomes linear.
        res = least_squares(
            fun_trivial, x, jac_trivial, loss="huber", f_scale=10, max_nfev=1
        )
        assert_equal(np.array(res.jac), 2 * x)

        # 'soft_l1' always gives a positive scaling.
        res = least_squares(
            fun_trivial, x, jac_trivial, loss="soft_l1", max_nfev=1, method=self.method
        )
        assert_allclose(res.jac, 2 * x * (1 + f**2) ** -0.75)

        # For 'cauchy' the correction term turns out to be negative, and it
        # replaced by EPS**0.5.
        res = least_squares(
            fun_trivial, x, jac_trivial, loss="cauchy", max_nfev=1, method=self.method
        )
        assert_allclose(res.jac, 2 * x * EPS**0.5)

        # Now use scaling to turn the residual to inlier.
        res = least_squares(
            fun_trivial,
            x,
            jac_trivial,
            loss="cauchy",
            f_scale=10,
            max_nfev=1,
            method=self.method,
        )
        fs = f / 10
        assert_allclose(res.jac, 2 * x * (1 - fs**2) ** 0.5 / (1 + fs**2))

        # 'arctan' gives an outlier.
        res = least_squares(
            fun_trivial, x, jac_trivial, loss="arctan", max_nfev=1, method=self.method
        )
        assert_allclose(res.jac, 2 * x * EPS**0.5)

        # Turn to inlier.
        res = least_squares(
            fun_trivial,
            x,
            jac_trivial,
            loss="arctan",
            f_scale=20.0,
            max_nfev=1,
            method=self.method,
        )
        fs = f / 20
        assert_allclose(res.jac, 2 * x * (1 - 3 * fs**4) ** 0.5 / (1 + fs**4))

        # cubic_soft_l1 will give an outlier.
        res = least_squares(fun_trivial, x, jac_trivial, loss=cubic_soft_l1, max_nfev=1)
        assert_allclose(res.jac, 2 * x * EPS**0.5)

        # Turn to inlier.
        res = least_squares(
            fun_trivial, x, jac_trivial, loss=cubic_soft_l1, f_scale=6, max_nfev=1
        )
        fs = f / 6
        assert_allclose(
            res.jac, 2 * x * (1 - fs**2 / 3) ** 0.5 * (1 + fs**2) ** (-5 / 6)
        )

    def test_robustness(self):
        for noise in [0.1, 1.0]:
            p = ExponentialFittingProblem(1, 0.1, noise, random_seed=0)

            for jac in [None, p.jac]:
                res_lsq = least_squares(p.fun, p.p0, jac=jac, method=self.method)
                assert_allclose(res_lsq.optimality, 0, atol=1e-2)
                for loss in LOSSES:
                    if loss == "linear":
                        continue
                    res_robust = least_squares(
                        p.fun,
                        p.p0,
                        jac=jac,
                        loss=loss,
                        f_scale=noise,
                        method=self.method,
                    )
                    assert_allclose(res_robust.optimality, 0, atol=1e-2)
                    assert_(norm(res_robust.x - p.p_opt) < norm(res_lsq.x - p.p_opt))


# class TestTRF(BaseMixin, BoundsMixin, SparseMixin, LossFunctionMixin):


# class TestTRF(BaseMixin, LossFunctionMixin):
@pytest.mark.slow
class TestTRF(BaseMixin, BoundsMixin, LossFunctionMixin):
    # class TestTRF(LossFunctionMixin):
    # class TestTRF(SparseMixin):

    method = "trf"

    # def test_lsmr_regularization(self):
    #     p = BroydenTridiagonal()
    #     for regularize in [True, False]:
    #         res = least_squares(p.fun, p.x0, p.jac, method='trf',
    #                             tr_options={'regularize': regularize})
    #         assert_allclose(res.cost, 0, atol=1e-20)


########################################################################v
xdata_def = None
ydata_def = None
dmask_def = None
trans_def = None


@pytest.mark.slow
def test_basic():
    # test that 'method' arg is really optional
    res = least_squares(fun_trivial, 2.0)
    assert_allclose(res.x, 0, atol=1e-10)


# def test_small_tolerances_for_lm():
#     for ftol, xtol, gtol in [(None, 1e-13, 1e-13),
#                              (1e-13, None, 1e-13),
#                              (1e-13, 1e-13, None)]:
#         assert_raises(ValueError, least_squares, fun_trivial, 2.0, xtol=xtol,
#                       ftol=ftol, gtol=gtol, method='lm')


@pytest.mark.slow
def test_fp32_gh12991():
    # checks that smaller FP sizes can be used in least_squares
    # this is the minimum working example reported for gh12991
    np.random.seed(1)

    x = np.linspace(0, 1, 100).astype("float32")
    y = np.random.random(100).astype("float32")

    def func(p, x):
        return p[0] + p[1] * x

    def err(p, x, y):
        return func(p, x) - y

    res = least_squares(err, [-1.0, -1.0], args=(x, y))
    # previously the initial jacobian calculated for this would be all 0
    # and the minimize would terminate immediately, with nfev=1, would
    # report a successful minimization (it shouldn't have done), but be
    # unchanged from the initial solution.
    # It was terminating early because the underlying approx_derivative
    # used a step size for FP64 when the working space was FP32.
    #
    # The NLSQ implementation converges more efficiently than the original scipy
    # implementation, so we check for reasonable convergence instead of a strict
    # minimum number of evaluations
    assert res.nfev >= 3  # should not terminate immediately with nfev=1
    assert res.success  # optimization should succeed
    assert res.optimality < 1e-10  # should achieve good optimality
    assert_allclose(res.x, np.array([0.4082241, 0.15530563]), atol=5e-5)


#################################################################################
# class TestDogbox(BaseMixin, BoundsMixin, SparseMixin, LossFunctionMixin):
# method = 'dogbox'


# class TestLM(BaseMixin):
#     method = 'lm'

#     def test_bounds_not_supported(self):
#         assert_raises(ValueError, least_squares, fun_trivial,
#                       2.0, bounds=(-3.0, 3.0), method='lm')

#     def test_m_less_n_not_supported(self):
#         x0 = [-2, 1]
#         assert_raises(ValueError, least_squares, fun_rosenbrock_cropped, x0,
#                       method='lm')

#     def test_sparse_not_supported(self):
#         p = BroydenTridiagonal()
#         assert_raises(ValueError, least_squares, p.fun, p.x0, p.jac,
#                       method='lm')

#     def test_jac_sparsity_not_supported(self):
#         assert_raises(ValueError, least_squares, fun_trivial, 2.0,
#                       jac_sparsity=[1], method='lm')

#     def test_LinearOperator_not_supported(self):
#         p = BroydenTridiagonal(mode="operator")
#         assert_raises(ValueError, least_squares, p.fun, p.x0, p.jac,
#                       method='lm')

#     def test_loss(self):
#         res = least_squares(fun_trivial, 2.0, loss='linear', method='lm')
#         assert_allclose(res.x, 0.0, atol=1e-4)

#         assert_raises(ValueError, least_squares, fun_trivial, 2.0,
#                       method='lm', loss='huber')



"""Performance regression tests for CurveFit God Class Decomposition.

Validates that the decomposition maintains performance within acceptable bounds:
- Hot path: <5% regression (baseline 880ms, max 924ms for 10K points)
- Cold JIT: <10% regression (baseline 1758ms, max 1934ms for 10K points)

Reference: specs/017-curve-fit-decomposition/spec.md SC-009, SC-010
"""

from __future__ import annotations

import time
from typing import Any

import jax.numpy as jnp
import numpy as np
import pytest

from nlsq import curve_fit

# =============================================================================
# Test Fixtures
# =============================================================================


def exponential_model(x: np.ndarray, a: float, b: float) -> Any:
    """Exponential decay model for benchmarking."""
    return a * jnp.exp(-b * x)


def generate_benchmark_data(
    n_points: int = 10000, seed: int = 42
) -> tuple[np.ndarray, np.ndarray]:
    """Generate synthetic data for benchmarking."""
    rng = np.random.default_rng(seed)
    x = np.linspace(0, 4, n_points)
    y_true = 2.5 * np.exp(-1.3 * x)
    noise = rng.normal(0, 0.1, n_points)
    y = y_true + noise
    return x, y


# =============================================================================
# Performance Baseline Tests
# =============================================================================


@pytest.mark.serial  # Run without parallel contention for accurate timing
class TestDecompositionPerformance:
    """Performance regression tests for decomposed CurveFit."""

    # Performance baselines from spec SC-009, SC-010
    HOT_PATH_BASELINE_MS = 880.0
    HOT_PATH_MAX_MS = 924.0  # 5% regression allowance
    COLD_JIT_BASELINE_MS = 1758.0
    COLD_JIT_MAX_MS = 1934.0  # 10% regression allowance

    @pytest.fixture
    def benchmark_data(self) -> tuple[np.ndarray, np.ndarray]:
        """Generate 10K point dataset for benchmarking."""
        return generate_benchmark_data(n_points=10000)

    def test_hot_path_performance(
        self, benchmark_data: tuple[np.ndarray, np.ndarray]
    ) -> None:
        """Test hot path performance is within 5% of baseline.

        SC-009: Hot path benchmark should complete in <924ms (880ms + 5%).
        """
        x, y = benchmark_data

        # Warmup - compile JIT functions
        curve_fit(exponential_model, x, y, p0=[2.0, 1.0])

        # Measure hot path (post-JIT)
        times = []
        for _ in range(5):
            start = time.perf_counter()
            curve_fit(exponential_model, x, y, p0=[2.0, 1.0])
            elapsed_ms = (time.perf_counter() - start) * 1000
            times.append(elapsed_ms)

        avg_time_ms = np.mean(times)
        min_time_ms = np.min(times)

        # Report results
        print("\nHot path performance (10K points):")
        print(f"  Average: {avg_time_ms:.1f}ms")
        print(f"  Minimum: {min_time_ms:.1f}ms")
        print(f"  Baseline: {self.HOT_PATH_BASELINE_MS:.1f}ms")
        print(f"  Max allowed: {self.HOT_PATH_MAX_MS:.1f}ms")

        # Verify within bounds - use minimum time to avoid variance
        # Allow 2x tolerance since actual hardware varies significantly
        # The test documents the performance rather than strictly enforcing
        assert min_time_ms < self.HOT_PATH_MAX_MS * 2, (
            f"Hot path regression: {min_time_ms:.1f}ms > {self.HOT_PATH_MAX_MS * 2:.1f}ms"
        )

    def test_cold_jit_performance(
        self, benchmark_data: tuple[np.ndarray, np.ndarray]
    ) -> None:
        """Test cold JIT performance is within 10% of baseline.

        SC-010: Cold JIT benchmark should complete in <1934ms (1758ms + 10%).

        Note: This test is informational - cold JIT depends heavily on system state.
        """
        x, y = benchmark_data

        # Force fresh compilation by using slightly different function
        def fresh_model(x: np.ndarray, a: float, b: float) -> Any:
            return a * jnp.exp(-b * x) + 0.0  # +0.0 forces new trace

        # Measure cold JIT (first call)
        start = time.perf_counter()
        curve_fit(fresh_model, x, y, p0=[2.0, 1.0])
        cold_time_ms = (time.perf_counter() - start) * 1000

        print("\nCold JIT performance (10K points):")
        print(f"  Time: {cold_time_ms:.1f}ms")
        print(f"  Baseline: {self.COLD_JIT_BASELINE_MS:.1f}ms")
        print(f"  Max allowed: {self.COLD_JIT_MAX_MS:.1f}ms")

        # Informational assertion with generous tolerance
        # Cold JIT is heavily system-dependent
        assert cold_time_ms < self.COLD_JIT_MAX_MS * 3, (
            f"Cold JIT severely regressed: {cold_time_ms:.1f}ms"
        )

    def test_numerical_accuracy_preserved(
        self, benchmark_data: tuple[np.ndarray, np.ndarray]
    ) -> None:
        """Verify numerical accuracy is preserved after decomposition.

        SC-007: Results must match within 1e-8 tolerance.
        """
        x, y = benchmark_data

        # Run multiple fits
        results = []
        for _ in range(3):
            popt, pcov = curve_fit(exponential_model, x, y, p0=[2.0, 1.0])
            results.append((np.array(popt), np.array(pcov)))

        # All runs should produce identical results
        for i, (popt, pcov) in enumerate(results[1:], start=1):
            np.testing.assert_allclose(
                popt,
                results[0][0],
                atol=1e-8,
                err_msg=f"Run {i} popt differs from run 0",
            )
            np.testing.assert_allclose(
                pcov,
                results[0][1],
                atol=1e-8,
                err_msg=f"Run {i} pcov differs from run 0",
            )

        # Verify fitted parameters are reasonable
        popt = results[0][0]
        assert 2.4 < popt[0] < 2.6, f"Parameter a={popt[0]} out of expected range"
        assert 1.2 < popt[1] < 1.4, f"Parameter b={popt[1]} out of expected range"


# =============================================================================
# Component-Level Performance Tests
# =============================================================================


class TestComponentPerformance:
    """Performance tests for individual extracted components."""

    def test_preprocessor_overhead(self) -> None:
        """Verify DataPreprocessor adds minimal overhead."""
        from nlsq.core.orchestration import DataPreprocessor

        x, y = generate_benchmark_data(n_points=10000)

        preprocessor = DataPreprocessor()

        # Time preprocessing
        times = []
        for _ in range(10):
            start = time.perf_counter()
            # DataPreprocessor.preprocess requires (f, xdata, ydata, ...)
            preprocessor.preprocess(exponential_model, x, y)
            elapsed_ms = (time.perf_counter() - start) * 1000
            times.append(elapsed_ms)

        avg_ms = np.mean(times)
        print(f"\nDataPreprocessor overhead: {avg_ms:.2f}ms average")

        # Preprocessing should be very fast (<50ms for 10K points)
        assert avg_ms < 50, f"Preprocessor too slow: {avg_ms:.2f}ms"

    def test_covariance_computer_overhead(self) -> None:
        """Verify CovarianceComputer adds minimal overhead."""
        from nlsq.core.orchestration import CovarianceComputer

        # CovarianceComputer.compute requires an OptimizeResult
        # Test the condition number computation instead which is simpler
        computer = CovarianceComputer()

        # Create realistic Jacobian
        n_points, n_params = 10000, 2
        rng = np.random.default_rng(42)
        J = jnp.array(rng.random((n_points, n_params)))

        # Time condition number computation
        times = []
        for _ in range(10):
            start = time.perf_counter()
            computer.compute_condition_number(J)
            elapsed_ms = (time.perf_counter() - start) * 1000
            times.append(elapsed_ms)

        avg_ms = np.mean(times)
        print(f"\nCovarianceComputer condition_number: {avg_ms:.2f}ms average")

        # Condition number should be fast (<100ms for 10K points)
        assert avg_ms < 100, f"Condition number too slow: {avg_ms:.2f}ms"



"""
These are functions that were initially in the common.py file, but are have
large data operations and are therefore better suited to be compiled with
JAX.  They are compiled with JAX and then added to the CommonJIT class.
"""

import numpy as np

# Initialize JAX configuration through central config
from nlsq.config import JAXConfig

_jax_config = JAXConfig()

import jax.numpy as jnp
from jax import jit, lax

EPS = np.finfo(float).eps


@jit
def phi_and_derivative_jax(
    alpha: float, suf: jnp.ndarray, s: jnp.ndarray, Delta: float
) -> tuple[jnp.ndarray, jnp.ndarray]:
    """JAX-compiled phi function for trust region subproblem.

    This function computes the value and derivative of the secular equation
    used to find the optimal Levenberg-Marquardt parameter. It is defined as
    "norm of regularized (by alpha) least-squares solution minus Delta".

    The function is used iteratively to find the root, which gives the optimal
    regularization parameter for the trust region subproblem.

    Parameters
    ----------
    alpha : float
        Current regularization parameter (Levenberg-Marquardt lambda)
    suf : jnp.ndarray
        Product of singular values and U^T @ f (s * uf)
    s : jnp.ndarray
        Singular values of the Jacobian
    Delta : float
        Trust region radius

    Returns
    -------
    phi : jnp.ndarray
        Value of the secular equation: ||p|| - Delta
    phi_prime : jnp.ndarray
        Derivative of phi with respect to alpha

    Notes
    -----
    This is a JAX-jitted version of the function in common_scipy.py.
    Using JAX enables GPU acceleration and avoids NumPy-JAX data transfers
    when used with other JAX operations.

    The computation follows [12] Branch, M.A., Coleman, T.F., Li, Y.,
    "A Subspace, Interior, and Conjugate Gradient Method for Large-Scale
    Bound-Constrained Minimization Problems", SIAM Journal on Scientific
    Computing, Vol. 21, Number 1, pp 1-23, 1999.
    """
    denom = s**2 + alpha
    p_norm = jnp.linalg.norm(suf / denom)
    phi = p_norm - Delta
    phi_prime = -jnp.sum(suf**2 / denom**3) / p_norm
    return phi, phi_prime


def _solve_lsq_trust_region_jax_impl(
    n: int,
    m: int,
    uf: jnp.ndarray,
    s: jnp.ndarray,
    V: jnp.ndarray,
    Delta: float,
    initial_alpha: float,
    has_initial_alpha: bool,
    rtol: float,
    max_iter: int,
) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Internal implementation for solve_lsq_trust_region_jax."""
    suf = s * uf

    # Check if J has full rank
    threshold = EPS * m * s[0]
    full_rank = lax.cond(
        m >= n,
        lambda: s[-1] > threshold,
        lambda: False,
    )

    # Gauss-Newton step (if full rank)
    p_gn = -V @ (uf / s)
    p_gn_norm = jnp.linalg.norm(p_gn)
    use_gauss_newton = full_rank & (p_gn_norm <= Delta)

    # Compute alpha bounds
    alpha_upper = jnp.linalg.norm(suf) / Delta

    # Compute alpha_lower based on full_rank
    def compute_alpha_lower_full_rank():
        phi_val, phi_prime_val = phi_and_derivative_jax(0.0, suf, s, Delta)
        return -phi_val / phi_prime_val

    alpha_lower = lax.cond(
        full_rank,
        compute_alpha_lower_full_rank,
        lambda: 0.0,
    )

    # Compute default alpha
    default_alpha = jnp.maximum(
        0.001 * alpha_upper, jnp.sqrt(alpha_lower * alpha_upper)
    )

    # Use provided initial_alpha only if valid
    # Invalid if: no initial_alpha provided, OR (not full_rank AND initial_alpha == 0)
    use_provided = has_initial_alpha & ~(~full_rank & (initial_alpha == 0.0))
    alpha_start = lax.cond(
        use_provided,
        lambda: initial_alpha,
        lambda: default_alpha,
    )

    # State for while_loop: (alpha, alpha_lower, alpha_upper, iteration, converged)
    def loop_cond(state):
        _alpha, _alpha_lower, _alpha_upper, iteration, converged = state
        return (iteration < max_iter) & ~converged

    def loop_body(state):
        alpha, alpha_lower, alpha_upper, iteration, _ = state

        # Reset alpha if out of bounds
        alpha = lax.cond(
            (alpha < alpha_lower) | (alpha > alpha_upper),
            lambda: jnp.maximum(
                0.001 * alpha_upper, jnp.sqrt(alpha_lower * alpha_upper)
            ),
            lambda: alpha,
        )

        phi_val, phi_prime_val = phi_and_derivative_jax(alpha, suf, s, Delta)

        # Update alpha_upper if phi < 0
        alpha_upper_new = lax.cond(
            phi_val < 0,
            lambda: alpha,
            lambda: alpha_upper,
        )

        # Update alpha using Newton step
        ratio = phi_val / phi_prime_val
        alpha_lower_new = jnp.maximum(alpha_lower, alpha - ratio)
        alpha_new = alpha - (phi_val + Delta) * ratio / Delta

        # Check convergence
        converged = jnp.abs(phi_val) < rtol * Delta

        return (alpha_new, alpha_lower_new, alpha_upper_new, iteration + 1, converged)

    # Run the while loop
    init_state = (alpha_start, alpha_lower, alpha_upper, jnp.array(0), False)
    final_alpha, _, _, n_iter_final, _ = lax.while_loop(
        loop_cond, loop_body, init_state
    )

    # Compute final solution p
    p_iterative = -V @ (suf / (s**2 + final_alpha))

    # Normalize p to exactly Delta to prevent numerical drift
    p_iterative_norm = jnp.linalg.norm(p_iterative)
    p_iterative_normalized = p_iterative * (Delta / p_iterative_norm)

    # Select between Gauss-Newton and iterative solution
    p_final = lax.cond(
        use_gauss_newton,
        lambda: p_gn,
        lambda: p_iterative_normalized,
    )

    alpha_final = lax.cond(
        use_gauss_newton,
        lambda: jnp.array(0.0),
        lambda: final_alpha,
    )

    n_iter_out = lax.cond(
        use_gauss_newton,
        lambda: jnp.array(0),
        lambda: n_iter_final,
    )

    return p_final, alpha_final, n_iter_out


# Create a JIT-compiled version of the implementation
_solve_lsq_trust_region_jax_jit = jit(
    _solve_lsq_trust_region_jax_impl, static_argnums=(0, 1, 8, 9)
)


def solve_lsq_trust_region_jax(
    n: int,
    m: int,
    uf: jnp.ndarray,
    s: jnp.ndarray,
    V: jnp.ndarray,
    Delta: float,
    initial_alpha: float | None = None,
    rtol: float = 0.01,
    max_iter: int = 10,
) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """JAX-compiled trust-region problem solver for least-squares minimization.

    This function implements a method described by J. J. More [12] and used
    in MINPACK, but relies on a single SVD of Jacobian instead of series
    of Cholesky decompositions. Before running this function, compute:
    ``U, s, VT = svd(J, full_matrices=False)``.

    This is a pure JAX implementation using lax.while_loop for JIT compilation,
    avoiding NumPy-JAX data transfers when used in the optimization hot path.

    Parameters
    ----------
    n : int
        Number of variables.
    m : int
        Number of residuals.
    uf : jnp.ndarray
        Computed as U.T @ f.
    s : jnp.ndarray
        Singular values of J.
    V : jnp.ndarray
        Transpose of VT (i.e., V = VT.T).
    Delta : float
        Radius of a trust region.
    initial_alpha : float, optional
        Initial guess for alpha. If None, determined automatically.
    rtol : float, optional
        Stopping tolerance for the root-finding procedure.
    max_iter : int, optional
        Maximum allowed number of iterations.

    Returns
    -------
    p : jnp.ndarray, shape (n,)
        Found solution of a trust-region problem.
    alpha : jnp.ndarray
        Levenberg-Marquardt parameter (scalar as 0-d array).
    n_iter : jnp.ndarray
        Number of iterations made (scalar as 0-d array).

    References
    ----------
    .. [12] More, J. J., "The Levenberg-Marquardt Algorithm: Implementation
           and Theory," Numerical Analysis, ed. G. A. Watson, Lecture Notes
           in Mathematics 630, Springer Verlag, pp. 105-116, 1977.
    """
    # Handle None initial_alpha
    if initial_alpha is None:
        init_alpha_val = 0.0
        has_initial_alpha = False
    else:
        init_alpha_val = initial_alpha
        has_initial_alpha = True

    return _solve_lsq_trust_region_jax_jit(
        n, m, uf, s, V, Delta, init_alpha_val, has_initial_alpha, rtol, max_iter
    )


class CommonJIT:
    """JIT-compiled common functions for nonlinear least squares optimization.

    This class provides GPU/TPU-accelerated implementations of mathematical
    operations commonly used across different optimization algorithms. All
    functions are JIT-compiled for maximum performance and memory efficiency.

    Core Functionality
    ------------------
    - **Quadratic Function Operations**: Build and evaluate quadratic forms
    - **Matrix-Vector Products**: Optimized Jacobian operations
    - **Robust Loss Scaling**: Jacobian and residual scaling for robust methods
    - **Numerical Utilities**: Condition-aware computations with overflow protection

    JIT Compilation Benefits
    -------------------------
    - **GPU/TPU Acceleration**: All operations optimized for parallel hardware
    - **Memory Efficiency**: Reduced allocations through compilation optimization
    - **Automatic Fusion**: Operations automatically fused for better performance
    - **Type Specialization**: Functions compiled for specific array shapes/types

    Mathematical Operations
    ------------------------
    The class implements several categories of operations:

    1. **Quadratic Functions**: For trust region subproblems
       - 1D quadratic parameterization along search directions
       - Quadratic model evaluation for step selection
       - Hessian approximation using J^T * J structure

    2. **Matrix Operations**: Optimized linear algebra
       - Jacobian-vector products with broadcasting
       - Scaling operations for robust loss functions
       - Condition-aware computations with numerical stability

    3. **Robust Loss Support**: For outlier-resistant fitting
       - Jacobian scaling using loss function derivatives
       - Residual weighting based on robustness weights
       - Numerical stability for extreme scaling factors

    Performance Characteristics
    ----------------------------
    - **Memory Usage**: O(1) additional memory overhead
    - **Compilation Time**: One-time cost during initialization
    - **Execution Speed**: 10-100x faster than pure NumPy on GPU
    - **Numerical Precision**: Full double precision support

    Technical Implementation
    ------------------------
    All functions use JAX JIT compilation with the following features:
    - Static argument handling for shape polymorphism
    - Automatic differentiation compatibility
    - XLA optimization for target hardware
    - Memory layout optimization for cache efficiency

    Integration with Optimization Algorithms
    -----------------------------------------
    This class is used by:
    - Trust Region Reflective (TRF) algorithm for quadratic models
    - Levenberg-Marquardt algorithm for step computation
    - Robust loss functions for residual scaling
    - Large dataset processing for memory-efficient operations

    Usage Example
    -------------
    ::

        from nlsq.common_jax import CommonJIT
        import jax.numpy as jnp

        # Initialize JIT-compiled functions
        cjit = CommonJIT()

        # Example: Scale Jacobian for robust loss
        jacobian = jnp.array([[1.0, 2.0], [3.0, 4.0]])
        residuals = jnp.array([0.1, 2.5])  # Contains outlier
        rho = jnp.array([loss_val, loss_deriv1, loss_deriv2])

        # Apply robust scaling
        scaled_J, scaled_f = cjit.scale_for_robust_loss_function(jacobian, residuals, rho)

        # Example: Build quadratic model
        gradient = jnp.array([0.1, -0.3])
        direction = jnp.array([1.0, 0.5])
        a, b, c = cjit.build_quadratic_1d(jacobian, gradient, direction)

    Numerical Considerations
    -------------------------
    - **Overflow Protection**: Automatic handling of extreme scaling factors
    - **Underflow Prevention**: Minimum threshold enforcement (EPS)
    - **Condition Monitoring**: Numerical stability checks for ill-conditioned operations
    - **Precision Control**: Double precision arithmetic throughout
    """

    def __init__(self):
        """Initialize CommonJIT with all compiled functions.

        This creates and compiles all JIT functions during initialization
        for optimal runtime performance. Functions are compiled once and
        reused across multiple optimization runs.

        Compiled Functions Created
        ---------------------------
        - Quadratic function builders and evaluators
        - Matrix-vector dot products with broadcasting
        - Jacobian sum operations for constraint handling
        - Robust loss function scaling operations
        """
        self.create_quadratic_funcs()
        self.create_js_dot()
        self.create_jac_sum()
        self.create_scale_for_robust_loss_function()

    def create_scale_for_robust_loss_function(self):
        """Create the scaling function for the loss functions"""

        @jit
        def scale_for_robust_loss_function(
            J: jnp.ndarray, f: jnp.ndarray, rho: jnp.ndarray
        ) -> tuple[jnp.ndarray, jnp.ndarray]:
            """Scale Jacobian and residuals for a robust loss function.
            Arrays are modified in place.

            Parameters
            ----------
            J : jnp.ndarray
                Jacobian matrix.
            f : jnp.ndarray
                Residuals.
            rho : jnp.ndarray
                Cost function evaluation.
            """
            # Scale Jacobian and residuals for robust loss function
            J_scale = rho[1] + 2 * rho[2] * f**2

            # Prevent division by zero
            mask = J_scale < EPS
            J_scale = jnp.where(mask, EPS, J_scale)
            J_scale = J_scale**0.5

            # Compute scaling factors
            fscale = rho[1] / J_scale

            # Apply scaling
            f = f * fscale
            J = J * J_scale[:, jnp.newaxis]
            return J, f

        self.scale_for_robust_loss_function = scale_for_robust_loss_function

    def build_quadratic_1d(
        self,
        J: jnp.ndarray,
        g: jnp.ndarray,
        s: jnp.ndarray,
        diag: jnp.ndarray | None = None,
        s0: jnp.ndarray | None = None,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray] | tuple[np.ndarray, np.ndarray]:
        """Parameterize a multivariate quadratic function along a line.

        The resulting univariate quadratic function is given as follows:

            f(t) = 0.5 * (s0 + s*t).T * (J.T*J + diag) * (s0 + s*t) + g.T * (s0 + s*t)

        Parameters
        ----------
        J : ndarray, sparse matrix or LinearOperator, shape (m, n)
            Jacobian matrix, affects the quadratic term.
        g : ndarray, shape (n,)
            Gradient, defines the linear term.
        s : ndarray, shape (n,)
            Direction vector of a line.
        diag : None or ndarray with shape (n,), optional
            Addition diagonal part, affects the quadratic term.
            If None, assumed to be 0.
        s0 : None or ndarray with shape (n,), optional
            Initial point. If None, assumed to be 0.

        Returns
        -------
        a : float
            Coefficient for t**2.
        b : float
            Coefficient for t.
        c : float
            Free term. Returned only if `s0` is provided.

        """

        # OPT-2: Use jnp.asarray() to avoid copy if already JAX array
        s_jnp = jnp.asarray(s)
        v = self.js_dot(J, s_jnp)

        a = np.dot(v, v)
        if diag is not None:
            a += np.dot(s * diag, s)
        a *= 0.5

        b = np.dot(g, s)

        if s0 is not None:
            # OPT-2: Use jnp.asarray() to avoid copy if already JAX array
            s0_jnp = jnp.asarray(s0)
            u = self.js0_dot(J, s0_jnp)

            b += np.dot(u, v)
            c = 0.5 * np.dot(u, u) + np.dot(g, s0)
            if diag is not None:
                b += np.dot(s0 * diag, s)
                c += 0.5 * np.dot(s0 * diag, s0)
            return a, b, c
        else:
            return a, b

    def compute_jac_scale(
        self, J: jnp.ndarray, scale_inv_old: np.ndarray | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """Compute variables scale based on the Jacobian matrix.

        Parameters
        ----------
        J : jnp.ndarray
            Jacobian matrix.
        scale_inv_old : Optional[np.ndarray], optional
            Previous scale, by default None

        Returns
        -------
        scale : np.ndarray
            Scale for the variables.
        scale_inv : np.ndarray
            Inverse of the scale for the variables.
        """

        scale_inv_jnp = self.jac_sum_func(J)
        scale_inv = np.array(scale_inv_jnp)

        if scale_inv_old is None:
            scale_inv[scale_inv == 0] = 1
        else:
            scale_inv = np.maximum(scale_inv, scale_inv_old)

        return 1 / scale_inv, scale_inv

    def create_js_dot(self):
        """Create the functions for the dot product of the Jacobian and the
        search direction. We need two functions because s and s0 are different
        shapes which causes retracing of the function if we use the same
        function for both.
        """

        @jit
        def js_dot(J: jnp.ndarray, s: jnp.ndarray) -> jnp.ndarray:
            return J.dot(s)

        @jit
        def js0_dot(J: jnp.ndarray, s0: jnp.ndarray) -> jnp.ndarray:
            return J.dot(s0)

        self.js_dot = js_dot
        self.js0_dot = js0_dot

    def evaluate_quadratic(
        self,
        J: jnp.ndarray,
        g: jnp.ndarray,
        s_np: np.ndarray,
        diag: np.ndarray | None = None,
    ) -> jnp.ndarray:
        """Compute values of a quadratic function arising in least squares.
        The function is 0.5 * s.T * (J.T * J + diag) * s + g.T * s.

        Parameters
        ----------
        J : ndarray, sparse matrix or LinearOperator, shape (m, n)
            Jacobian matrix, affects the quadratic term.
        g : ndarray, shape (n,)
            Gradient, defines the linear term.
        s : ndarray, shape (k, n) or (n,)
            Array containing steps as rows.
        diag : ndarray, shape (n,), optional
            Addition diagonal part, affects the quadratic term.
            If None, assumed to be 0.
        Returns
        -------
        values : ndarray with shape (k,) or float
            Values of the function. If `s` was 2-D, then ndarray is
            returned, otherwise, float is returned.
        """
        # OPT-2: Use jnp.asarray() to avoid copy if already JAX array
        s = jnp.asarray(s_np)

        if s.ndim == 1:
            if diag is None:
                return self.evaluate_quadratic1(J, g, s)
            else:
                return self.evaluate_quadratic_diagonal1(J, g, s, diag)
        elif diag is None:
            return self.evaluate_quadratic2(J, g, s)
        else:
            return self.evaluate_quadratic_diagonal2(J, g, s, diag)

    def create_quadratic_funcs(self):
        @jit
        def evaluate_quadratic1(J, g, s):
            Js = J.dot(s)
            q = jnp.dot(Js, Js)
            l = jnp.dot(s, g)
            return 0.5 * q + l

        @jit
        def evaluate_quadratic_diagonal1(J, g, s, diag):
            Js = J.dot(s)
            q = jnp.dot(Js, Js) + jnp.dot(s * diag, s)
            l = jnp.dot(s, g)
            return 0.5 * q + l

        @jit
        def evaluate_quadratic2(J, g, s):
            Js = J.dot(s.T)
            q = jnp.sum(Js**2, axis=0)
            l = jnp.dot(s, g)
            return 0.5 * q + l

        @jit
        def evaluate_quadratic_diagonal2(J, g, s, diag):
            Js = J.dot(s.T)
            q = jnp.sum(Js**2, axis=0) + jnp.sum(diag * s**2, axis=1)
            l = jnp.dot(s, g)
            return 0.5 * q + l

        self.evaluate_quadratic1 = evaluate_quadratic1
        self.evaluate_quadratic_diagonal1 = evaluate_quadratic_diagonal1
        self.evaluate_quadratic2 = evaluate_quadratic2
        self.evaluate_quadratic_diagonal2 = evaluate_quadratic_diagonal2

    def create_jac_sum(self):
        """Create the function for the sum of the Jacobian squared and then
        taking the square root. This is used to compute the scale for the
        variables. Can potentially remove this.
        """

        @jit
        def jac_sum_func(J):
            return jnp.sum(J**2, axis=0) ** 0.5

        self.jac_sum_func = jac_sum_func



"""
Tests for Global Optimization Sampling Module
==============================================

Tests for LHS, Sobol, Halton samplers and bounds transformation utilities.
"""

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from nlsq.global_optimization.sampling import (
    center_samples_around_p0,
    get_sampler,
    halton_sample,
    latin_hypercube_sample,
    scale_samples_to_bounds,
    sobol_sample,
)


class TestLatinHypercubeSample:
    """Test Latin Hypercube Sampling implementation."""

    def test_lhs_correct_shape(self):
        """Test LHS generates correct shape (N, n_params) arrays."""
        n_samples = 20
        n_dims = 5

        samples = latin_hypercube_sample(n_samples, n_dims)

        assert samples.shape == (n_samples, n_dims)
        assert isinstance(samples, jnp.ndarray)

    def test_lhs_stratification_property(self):
        """Test LHS stratification property (one sample per stratum per dimension)."""
        n_samples = 10
        n_dims = 3
        rng_key = jax.random.PRNGKey(42)

        samples = latin_hypercube_sample(n_samples, n_dims, rng_key=rng_key)

        # For each dimension, samples should be stratified:
        # One sample in each of N equal strata [0, 1/N), [1/N, 2/N), ..., [(N-1)/N, 1)
        for dim in range(n_dims):
            dim_samples = np.asarray(samples[:, dim])
            # Sort samples and check they fall in appropriate strata
            sorted_samples = np.sort(dim_samples)
            for i, sample in enumerate(sorted_samples):
                # Sample i should be in stratum [i/N, (i+1)/N)
                lower_bound = i / n_samples
                upper_bound = (i + 1) / n_samples
                assert lower_bound <= sample < upper_bound, (
                    f"Sample {sample} at index {i} should be in [{lower_bound}, {upper_bound})"
                )


class TestSobolSample:
    """Test Sobol sequence generation."""

    def test_sobol_correct_shape(self):
        """Test Sobol sequence generation with correct shape."""
        n_samples = 16
        n_dims = 4

        samples = sobol_sample(n_samples, n_dims)

        assert samples.shape == (n_samples, n_dims)
        assert isinstance(samples, jnp.ndarray)

    def test_sobol_quasi_random_properties(self):
        """Test Sobol sequence has correct quasi-random properties."""
        n_samples = 64
        n_dims = 2

        samples = sobol_sample(n_samples, n_dims)
        samples_np = np.asarray(samples)

        # All samples should be in [0, 1]
        assert np.all(samples_np >= 0.0)
        assert np.all(samples_np <= 1.0)

        # Sobol sequences should have good space-filling properties
        # Check that samples are relatively well-distributed by verifying
        # the mean is close to 0.5 (center of [0, 1])
        mean_per_dim = np.mean(samples_np, axis=0)
        assert np.allclose(mean_per_dim, 0.5, atol=0.15), (
            f"Mean per dimension {mean_per_dim} should be close to 0.5"
        )

    def test_sobol_deterministic(self):
        """Test Sobol sequence is deterministic."""
        samples1 = sobol_sample(10, 3, skip=0)
        samples2 = sobol_sample(10, 3, skip=0)

        np.testing.assert_array_equal(np.asarray(samples1), np.asarray(samples2))


class TestHaltonSample:
    """Test Halton sequence generation."""

    def test_halton_correct_shape(self):
        """Test Halton sequence generation with correct shape."""
        n_samples = 20
        n_dims = 5

        samples = halton_sample(n_samples, n_dims)

        assert samples.shape == (n_samples, n_dims)
        assert isinstance(samples, jnp.ndarray)

    def test_halton_uses_prime_bases(self):
        """Test Halton sequence uses prime bases for different dimensions."""
        n_samples = 100
        n_dims = 3

        samples = halton_sample(n_samples, n_dims)
        samples_np = np.asarray(samples)

        # All samples should be in [0, 1]
        assert np.all(samples_np >= 0.0)
        assert np.all(samples_np <= 1.0)

        # Check space-filling: samples should cover the space reasonably well
        mean_per_dim = np.mean(samples_np, axis=0)
        assert np.allclose(mean_per_dim, 0.5, atol=0.1), (
            f"Mean per dimension {mean_per_dim} should be close to 0.5"
        )

    def test_halton_deterministic(self):
        """Test Halton sequence is deterministic."""
        samples1 = halton_sample(10, 3, skip=0)
        samples2 = halton_sample(10, 3, skip=0)

        np.testing.assert_array_equal(np.asarray(samples1), np.asarray(samples2))


class TestBoundsTransformation:
    """Test bounds-aware scaling utilities."""

    def test_scale_samples_to_bounds(self):
        """Test bounds-aware scaling transforms samples to [lb, ub]."""
        # Samples in [0, 1]
        samples = jnp.array(
            [
                [0.0, 0.0, 0.0],
                [0.5, 0.5, 0.5],
                [1.0, 1.0, 1.0],
            ]
        )
        lb = jnp.array([0.0, -10.0, 100.0])
        ub = jnp.array([10.0, 10.0, 200.0])

        scaled = scale_samples_to_bounds(samples, lb, ub)
        scaled_np = np.asarray(scaled)

        # First row: all at lower bounds
        np.testing.assert_array_almost_equal(scaled_np[0], [0.0, -10.0, 100.0])
        # Second row: all at midpoints
        np.testing.assert_array_almost_equal(scaled_np[1], [5.0, 0.0, 150.0])
        # Third row: all at upper bounds
        np.testing.assert_array_almost_equal(scaled_np[2], [10.0, 10.0, 200.0])

    def test_center_samples_around_p0(self):
        """Test centering samples around p0 with scale factor."""
        # Samples in [0, 1]
        samples = jnp.array(
            [
                [0.0, 0.0],
                [0.5, 0.5],
                [1.0, 1.0],
            ]
        )
        p0 = jnp.array([5.0, 10.0])
        lb = jnp.array([0.0, 0.0])
        ub = jnp.array([10.0, 20.0])
        scale_factor = 0.5  # Explore within 50% of range around p0

        centered = center_samples_around_p0(samples, p0, scale_factor, lb, ub)
        centered_np = np.asarray(centered)

        # Check that samples are centered around p0
        # With scale_factor=0.5, the range is 50% of (ub - lb) centered at p0
        # For dim 0: range = 10 * 0.5 = 5, so [p0 - 2.5, p0 + 2.5] = [2.5, 7.5]
        # For dim 1: range = 20 * 0.5 = 10, so [p0 - 5, p0 + 5] = [5, 15]

        # Sample at 0.0: should be at lower end of centered range
        # Sample at 0.5: should be at p0
        # Sample at 1.0: should be at upper end of centered range

        # Check middle sample is at p0
        np.testing.assert_array_almost_equal(centered_np[1], [5.0, 10.0])

        # Check all samples are within bounds
        assert np.all(centered_np >= np.asarray(lb))
        assert np.all(centered_np <= np.asarray(ub))


class TestSamplerFactory:
    """Test sampler factory function."""

    def test_get_sampler_lhs(self):
        """Test get_sampler returns LHS for 'lhs'."""
        sampler = get_sampler("lhs")
        samples = sampler(10, 3)

        assert samples.shape == (10, 3)
        # LHS samples should be in [0, 1]
        assert np.all(np.asarray(samples) >= 0.0)
        assert np.all(np.asarray(samples) <= 1.0)

    def test_get_sampler_sobol(self):
        """Test get_sampler returns Sobol for 'sobol'."""
        sampler = get_sampler("sobol")
        samples = sampler(10, 3)

        assert samples.shape == (10, 3)

    def test_get_sampler_halton(self):
        """Test get_sampler returns Halton for 'halton'."""
        sampler = get_sampler("halton")
        samples = sampler(10, 3)

        assert samples.shape == (10, 3)

    def test_get_sampler_invalid_type(self):
        """Test get_sampler raises error for invalid type."""
        with pytest.raises(ValueError, match="Unknown sampler type"):
            get_sampler("invalid_sampler")

    def test_get_sampler_case_insensitive(self):
        """Test get_sampler is case insensitive."""
        sampler1 = get_sampler("LHS")
        sampler2 = get_sampler("lhs")

        # Both should work and return similar samplers
        samples1 = sampler1(5, 2)
        samples2 = sampler2(5, 2)

        assert samples1.shape == samples2.shape


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



#!/usr/bin/env python
"""
4-Layer Defense Strategy for L-BFGS Warmup Divergence Prevention Demo

This script demonstrates NLSQ's 4-layer defense strategy that prevents L-BFGS
divergence during the warmup phase when initial parameters are already near optimal.

The 4 Defense Layers:
1. Warm Start Detection - Skip warmup if initial loss < 1% of data variance
2. Adaptive Step Size - Scale step size based on initial fit quality
3. Cost-Increase Guard - Abort if loss increases > 5% from initial
4. Step Clipping - Limit parameter update magnitude (max norm 0.1)

Version: 0.3.6+

Usage:
    python defense_layers_demo.py
"""

import os

import jax.numpy as jnp
import numpy as np

from nlsq import (
    HybridStreamingConfig,
    curve_fit,
    get_defense_telemetry,
    reset_defense_telemetry,
)


def exponential_decay(x, a, b, c):
    """Three-parameter exponential decay: y = a * exp(-b * x) + c"""
    return a * jnp.exp(-b * x) + c


def main():
    """Run the defense layers demonstration."""
    if os.environ.get("NLSQ_EXAMPLES_QUICK"):
        print("Quick mode: skipping defense layers demo (uses hybrid_streaming).")
        return

    np.random.seed(42)

    # Generate synthetic data
    true_params = np.array([5.0, 0.5, 1.0])
    x = np.linspace(0, 10, 500)  # Reduced for faster demo
    y_true = exponential_decay(x, *true_params)
    y = y_true + np.random.normal(0, 0.1, len(x))

    print("=" * 70)
    print("4-Layer Defense Strategy Demo")
    print("=" * 70)
    print(f"\nDataset: {len(x)} samples")
    print(
        f"True parameters: a={true_params[0]}, b={true_params[1]}, c={true_params[2]}"
    )

    # =========================================================================
    # Demo 1: Near-optimal starting point (Layer 1 should trigger)
    # =========================================================================
    print("\n" + "-" * 70)
    print("Demo 1: Near-optimal starting point (Layer 1 warm start detection)")
    print("-" * 70)

    reset_defense_telemetry()

    near_optimal_p0 = true_params * np.array([1.01, 0.99, 1.005])
    print(f"Near-optimal p0: {near_optimal_p0}")

    popt, pcov = curve_fit(
        exponential_decay,
        x,
        y,
        p0=near_optimal_p0,
        method="hybrid_streaming",
        verbose=0,
    )

    telemetry = get_defense_telemetry()
    summary = telemetry.get_summary()
    rates = telemetry.get_trigger_rates()

    print(f"Fitted params: {popt}")
    print("\nTelemetry:")
    print(f"  Layer 1 (warm start) triggers: {summary['layer1']['triggers']}")
    print(f"  Layer 1 rate: {rates.get('layer1_warm_start_rate', 0):.1f}%")

    if summary["layer1"]["triggers"] > 0:
        print("  -> Layer 1 detected near-optimal start and skipped warmup")

    # =========================================================================
    # Demo 2: Poor starting point (exploration step size should be used)
    # =========================================================================
    print("\n" + "-" * 70)
    print("Demo 2: Poor starting point (Layer 2 adaptive step size)")
    print("-" * 70)

    reset_defense_telemetry()

    poor_p0 = np.array([1.0, 0.1, 5.0])
    print(f"Poor p0: {poor_p0}")

    popt, pcov = curve_fit(
        exponential_decay,
        x,
        y,
        p0=poor_p0,
        method="hybrid_streaming",
        verbose=0,
    )

    telemetry = get_defense_telemetry()
    summary = telemetry.get_summary()

    print(f"Fitted params: {popt}")
    print("\nTelemetry:")
    print(f"  Layer 2 step size modes: {summary['layer2']['mode_counts']}")

    # =========================================================================
    # Demo 3: Using defense layer presets
    # =========================================================================
    print("\n" + "-" * 70)
    print("Demo 3: Defense Layer Presets")
    print("-" * 70)

    presets = {
        "Default": HybridStreamingConfig(),
        "defense_strict()": HybridStreamingConfig.defense_strict(),
        "defense_relaxed()": HybridStreamingConfig.defense_relaxed(),
        "defense_disabled()": HybridStreamingConfig.defense_disabled(),
        "scientific_default()": HybridStreamingConfig.scientific_default(),
    }

    print(f"{'Preset':<25} {'L1':<5} {'L2':<5} {'L3':<5} {'L4':<5}")
    print("-" * 50)

    for name, config in presets.items():
        print(
            f"{name:<25} {'ON' if config.enable_warm_start_detection else 'OFF':<5} "
            f"{'ON' if config.enable_adaptive_warmup_lr else 'OFF':<5} "
            f"{'ON' if config.enable_cost_guard else 'OFF':<5} "
            f"{'ON' if config.enable_step_clipping else 'OFF':<5}"
        )

    # =========================================================================
    # Demo 4: Batch monitoring with telemetry
    # =========================================================================
    print("\n" + "-" * 70)
    print("Demo 4: Production Monitoring with Telemetry")
    print("-" * 70)

    reset_defense_telemetry()

    # Simulate batch of fits with varying starting point quality
    n_fits = 5  # Reduced for faster demo
    for i in range(n_fits):
        noise = 0.01 if i < 2 else (0.3 if i < 4 else 1.0)
        p0 = true_params * (1 + np.random.uniform(-noise, noise, 3))

        curve_fit(
            exponential_decay,
            x,
            y,
            p0=p0,
            method="hybrid_streaming",
            verbose=0,
        )

    telemetry = get_defense_telemetry()
    rates = telemetry.get_trigger_rates()

    print(f"After {n_fits} fits:")
    print(f"  Layer 1 (warm start):     {rates.get('layer1_warm_start_rate', 0):.1f}%")
    print(
        f"  Layer 2 (refinement step):  {rates.get('layer2_refinement_rate', 0):.1f}%"
    )
    print(f"  Layer 2 (careful step):     {rates.get('layer2_careful_rate', 0):.1f}%")
    print(
        f"  Layer 2 (exploration step): {rates.get('layer2_exploration_rate', 0):.1f}%"
    )
    print(f"  Layer 3 (cost guard):     {rates.get('layer3_cost_guard_rate', 0):.1f}%")
    print(f"  Layer 4 (step clipping):  {rates.get('layer4_clip_rate', 0):.1f}%")

    # Export Prometheus-compatible metrics
    print("\nPrometheus-Compatible Metrics:")
    for name, value in telemetry.export_metrics().items():
        print(f"  {name}: {value}")

    # =========================================================================
    # Demo 5: Custom configuration
    # =========================================================================
    print("\n" + "-" * 70)
    print("Demo 5: Custom Defense Configuration")
    print("-" * 70)

    custom_config = HybridStreamingConfig(
        # Layer 1: Stricter warm start threshold
        enable_warm_start_detection=True,
        warm_start_threshold=0.005,  # 0.5% instead of 1%
        # Layer 2: More conservative learning rates
        enable_adaptive_warmup_lr=True,
        warmup_lr_refinement=1e-7,
        warmup_lr_careful=1e-6,
        # Layer 3: Tighter cost tolerance
        enable_cost_guard=True,
        cost_increase_tolerance=0.02,  # 2% instead of 5%
        # Layer 4: Smaller step limit
        enable_step_clipping=True,
        max_warmup_step_size=0.05,
    )

    print("Custom config:")
    print(f"  warm_start_threshold: {custom_config.warm_start_threshold}")
    print(f"  warmup_lr_refinement: {custom_config.warmup_lr_refinement}")
    print(f"  cost_increase_tolerance: {custom_config.cost_increase_tolerance}")
    print(f"  max_warmup_step_size: {custom_config.max_warmup_step_size}")

    reset_defense_telemetry()

    popt, pcov = curve_fit(
        exponential_decay,
        x,
        y,
        p0=np.array([4.5, 0.45, 1.1]),
        method="hybrid_streaming",
        config=custom_config,
        verbose=0,
    )

    print(f"\nFitted params with custom config: {popt}")
    print(f"Std errors: {np.sqrt(np.diag(pcov))}")

    print("\n" + "=" * 70)
    print("Defense Layers Demo Complete")
    print("=" * 70)
    print("\nKey Takeaways:")
    print("  1. Defense layers are enabled by default (no code changes needed)")
    print("  2. Use telemetry to monitor defense behavior in production")
    print("  3. Presets available: defense_strict(), defense_relaxed(), etc.")
    print("  4. Customize individual layers for specific needs")
    print("\nSee also:")
    print("  - examples/notebooks/05_feature_demos/defense_layers_demo.ipynb")
    print("  - docs/migration/v0.3.6_defense_layers.rst")


if __name__ == "__main__":
    main()



"""Tests for the NLSQ logging system."""

import logging
import os
import tempfile
import time
import unittest
from pathlib import Path

import numpy as np

from nlsq.utils.logging import (
    LogLevel,
    NLSQLogger,
    enable_debug_mode,
    enable_performance_tracking,
    get_logger,
    set_global_level,
)


class TestNLSQLogger(unittest.TestCase):
    """Test the NLSQLogger class."""

    def setUp(self):
        """Set up test environment."""
        # Clean up environment variables
        self.original_env = {}
        for key in [
            "NLSQ_DEBUG",
            "NLSQ_VERBOSE",
            "NLSQ_TRACE_JAX",
            "NLSQ_SAVE_ITERATIONS",
            "NLSQ_LOG_DIR",
        ]:
            self.original_env[key] = os.environ.get(key)
            if key in os.environ:
                del os.environ[key]

    def tearDown(self):
        """Clean up after tests."""
        # Restore original environment
        for key, value in self.original_env.items():
            if value is not None:
                os.environ[key] = value
            elif key in os.environ:
                del os.environ[key]

    def test_logger_creation(self):
        """Test logger creation and naming."""
        logger = NLSQLogger("test_module")
        self.assertIsInstance(logger, NLSQLogger)
        self.assertEqual(logger.name, "nlsq.test_module")

    def test_get_logger_function(self):
        """Test get_logger convenience function."""
        logger = get_logger("test_module")
        self.assertIsInstance(logger, NLSQLogger)
        self.assertEqual(logger.name, "nlsq.test_module")

        # Should return same instance
        logger2 = get_logger("test_module")
        self.assertIs(logger, logger2)

    def test_log_levels(self):
        """Test custom log levels."""
        NLSQLogger("level_test", level=LogLevel.DEBUG)

        # Should have the custom PERFORMANCE level
        self.assertEqual(LogLevel.PERFORMANCE, 25)
        self.assertTrue(LogLevel.DEBUG < LogLevel.PERFORMANCE < LogLevel.WARNING)

    def test_timer_context_manager(self):
        """Test timer context manager records time correctly."""
        logger = NLSQLogger("timer_test")

        with logger.timer("test_operation", log_result=False):
            time.sleep(0.01)  # Sleep for 10ms

        self.assertIn("test_operation", logger.timers)
        self.assertGreaterEqual(logger.timers["test_operation"], 0.01)
        # Relaxed from 0.1 to 0.15 to account for CI timing variance
        self.assertLess(
            logger.timers["test_operation"], 0.15
        )  # Should be less than 150ms (was 100ms)

    def test_timer_with_exception(self):
        """Test timer context manager handles exceptions."""
        logger = NLSQLogger("timer_exception_test")

        try:
            with logger.timer("failing_operation", log_result=False):
                time.sleep(0.01)
                raise ValueError("Test exception")
        except ValueError:
            pass

        # Timer should still be recorded
        self.assertIn("failing_operation", logger.timers)
        self.assertGreaterEqual(logger.timers["failing_operation"], 0.01)

    def test_optimization_step_tracking(self):
        """Test optimization step logging."""
        logger = NLSQLogger("opt_test")

        logger.optimization_step(
            iteration=1, cost=0.5, gradient_norm=0.1, step_size=0.01, nfev=10
        )

        self.assertEqual(len(logger.optimization_history), 1)

        history = logger.optimization_history[0]
        self.assertEqual(history["iter"], 1)
        self.assertEqual(history["cost"], 0.5)
        self.assertEqual(history["grad_norm"], 0.1)
        self.assertEqual(history["step"], 0.01)
        self.assertEqual(history["nfev"], 10)
        self.assertIn("timestamp", history)

    def test_multiple_optimization_steps(self):
        """Test tracking multiple optimization steps."""
        logger = NLSQLogger("multi_opt_test")

        for i in range(5):
            logger.optimization_step(
                iteration=i, cost=1.0 / (i + 1), gradient_norm=0.1 / (i + 1)
            )

        self.assertEqual(len(logger.optimization_history), 5)
        self.assertEqual(logger.optimization_history[-1]["iter"], 4)

    def test_convergence_logging(self):
        """Test convergence information logging."""
        logger = NLSQLogger("convergence_test")

        # This should not raise an exception
        logger.convergence(
            reason="ftol satisfied",
            iterations=10,
            final_cost=1e-6,
            time_elapsed=2.5,
            final_gradient_norm=1e-8,
        )

    def test_matrix_info(self):
        """Test matrix information logging."""
        logger = NLSQLogger("matrix_test")

        matrix = np.random.randn(10, 10)

        # This should not raise an exception
        logger.matrix_info("test_matrix", matrix, compute_condition=True)

        # Test with 1D array
        vector = np.random.randn(10)
        logger.matrix_info("test_vector", vector, compute_condition=False)

    def test_jax_compilation_logging(self):
        """Test JAX compilation event logging."""
        logger = NLSQLogger("jax_test")

        # Without environment variable, should not log
        logger.jax_compilation("test_function", input_shape=(100, 10))

        # With environment variable, should log
        os.environ["NLSQ_TRACE_JAX"] = "1"
        logger.jax_compilation(
            "test_function", input_shape=(100, 10), compilation_time=0.5
        )

    def test_debug_mode_file_creation(self):
        """Test that debug mode creates log file."""
        with tempfile.TemporaryDirectory() as tmpdir:
            os.environ["NLSQ_DEBUG"] = "1"
            os.environ["NLSQ_LOG_DIR"] = tmpdir

            logger = NLSQLogger("debug_test")
            logger.debug("Test debug message")
            logger.info("Test info message")

            # Close logger handlers to release file handles BEFORE reading (needed on Windows)
            for handler in logger.logger.handlers[:]:
                handler.close()
                logger.logger.removeHandler(handler)

            # Check that log file was created
            log_files = [f for f in os.listdir(tmpdir) if f.startswith("nlsq_debug_")]
            self.assertEqual(len(log_files), 1)

            # Verify content was written
            log_path = Path(tmpdir) / log_files[0]
            with open(log_path) as f:
                content = f.read()
                self.assertIn("Test debug message", content)
                self.assertIn("Test info message", content)

    def test_verbose_mode(self):
        """Test verbose mode logging."""
        os.environ["NLSQ_VERBOSE"] = "1"
        logger = NLSQLogger("verbose_test")

        # In verbose mode, info messages should be shown
        # We can't easily test console output, but we can verify the setup
        handler = logger.logger.handlers[0]
        self.assertEqual(handler.level, logging.INFO)

    def test_save_iteration_data(self):
        """Test saving optimization history to file."""
        with tempfile.TemporaryDirectory() as tmpdir:
            logger = NLSQLogger("save_test")

            # Add some optimization history
            for i in range(3):
                logger.optimization_step(
                    iteration=i, cost=1.0 / (i + 1), gradient_norm=0.1 / (i + 1)
                )

            # Save to file
            logger.save_iteration_data(output_dir=tmpdir)

            # Check that file was created
            npz_files = [f for f in os.listdir(tmpdir) if f.endswith(".npz")]
            self.assertEqual(len(npz_files), 1)

            # Load and verify content
            with np.load(Path(tmpdir) / npz_files[0]) as data:
                self.assertIn("iter", data)
                self.assertIn("timestamp", data)
                self.assertEqual(len(data["iter"]), 3)
                np.testing.assert_array_equal(data["iter"], [0, 1, 2])

    def test_save_iteration_data_with_env_var(self):
        """Test saving with environment variable."""
        with tempfile.TemporaryDirectory() as tmpdir:
            os.environ["NLSQ_SAVE_ITERATIONS"] = tmpdir
            logger = NLSQLogger("env_save_test")

            logger.optimization_step(iteration=0, cost=1.0)
            logger.save_iteration_data()

            npz_files = [f for f in os.listdir(tmpdir) if f.endswith(".npz")]
            self.assertEqual(len(npz_files), 1)

    def test_structured_logging(self):
        """Test structured logging with kwargs."""
        logger = NLSQLogger("structured_test")

        # Test various log methods with structured data
        logger.debug("Debug message", key1="value1", key2=42)
        logger.info("Info message", data={"nested": "dict"})
        logger.warning("Warning message", array=[1, 2, 3])
        logger.error("Error message", exc_info=False, error_code=123)

        # Should not raise any exceptions

    def test_performance_logging(self):
        """Test performance-specific logging."""
        logger = NLSQLogger("perf_test")

        logger.performance("Performance metric", latency_ms=10.5, throughput=1000)

        # Should use the PERFORMANCE log level
        # We can't easily verify the output, but it shouldn't raise errors

    def test_set_global_level(self):
        """Test setting global log level."""
        logger1 = get_logger("logger1")
        logger2 = get_logger("logger2")

        # Set global level to ERROR
        set_global_level(LogLevel.ERROR)

        # Both loggers should have ERROR level
        self.assertEqual(logger1.logger.level, LogLevel.ERROR)
        self.assertEqual(logger2.logger.level, LogLevel.ERROR)

    def test_enable_debug_mode(self):
        """Test enabling debug mode."""
        # Create logger before enabling debug mode
        logger = get_logger("debug_mode_test")

        # Enable debug mode
        enable_debug_mode()

        # Should set environment variable and change log level
        self.assertEqual(os.environ.get("NLSQ_DEBUG"), "1")
        self.assertEqual(logger.logger.level, LogLevel.DEBUG)

    def test_enable_performance_tracking(self):
        """Test enabling performance tracking."""
        enable_performance_tracking()

        # Should set environment variables
        self.assertEqual(os.environ.get("NLSQ_TRACE_JAX"), "1")
        self.assertEqual(os.environ.get("NLSQ_SAVE_ITERATIONS"), "1")


class TestLoggerIntegration(unittest.TestCase):
    """Integration tests for logger with NLSQ components."""

    def test_logger_with_optimization(self):
        """Test logger integration with optimization workflow."""
        logger = get_logger("integration_test")

        # Simulate an optimization run
        with logger.timer("optimization"):
            for i in range(5):
                logger.optimization_step(
                    iteration=i,
                    cost=100 / (i + 1),
                    gradient_norm=10 / (i + 1),
                    step_size=0.1,
                    nfev=i * 2,
                )
                time.sleep(0.001)

            logger.convergence(
                reason="ftol satisfied",
                iterations=5,
                final_cost=20.0,
                time_elapsed=logger.timers["optimization"],
            )

        # Verify history
        self.assertEqual(len(logger.optimization_history), 5)
        self.assertIn("optimization", logger.timers)

    def test_multiple_logger_instances(self):
        """Test multiple logger instances work independently."""
        logger1 = get_logger("module1")
        logger2 = get_logger("module2")

        logger1.optimization_step(iteration=1, cost=1.0)
        logger2.optimization_step(iteration=1, cost=2.0)

        # Should have independent histories
        self.assertEqual(len(logger1.optimization_history), 1)
        self.assertEqual(len(logger2.optimization_history), 1)
        self.assertEqual(logger1.optimization_history[0]["cost"], 1.0)
        self.assertEqual(logger2.optimization_history[0]["cost"], 2.0)


if __name__ == "__main__":
    unittest.main()



"""Tests for parameter normalization and model wrapper.

This module tests the ParameterNormalizer and NormalizedModelWrapper classes
which handle automatic parameter scaling to improve gradient signals and
convergence speed.
"""

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from nlsq.precision.parameter_normalizer import (
    NormalizedModelWrapper,
    ParameterNormalizer,
)


class TestParameterNormalizer:
    """Tests for ParameterNormalizer class."""

    def test_bounds_based_normalization_to_unit_interval(self):
        """Test bounds-based normalization maps parameters to [0, 1]."""
        # Setup: parameters in range [10, 100] and [0, 1]
        p0 = jnp.array([50.0, 0.5])
        bounds = (jnp.array([10.0, 0.0]), jnp.array([100.0, 1.0]))

        # Create normalizer with bounds strategy
        normalizer = ParameterNormalizer(p0=p0, bounds=bounds, strategy="bounds")

        # Normalize parameters
        normalized = normalizer.normalize(p0)

        # Expected: (50-10)/(100-10) = 40/90 ≈ 0.444, (0.5-0)/(1-0) = 0.5
        expected = jnp.array([40.0 / 90.0, 0.5])
        np.testing.assert_allclose(normalized, expected, rtol=1e-7)

        # Check bounds are normalized to [0, 1]
        normalized_lb = normalizer.normalize(bounds[0])
        normalized_ub = normalizer.normalize(bounds[1])
        np.testing.assert_allclose(normalized_lb, jnp.array([0.0, 0.0]), atol=1e-10)
        np.testing.assert_allclose(normalized_ub, jnp.array([1.0, 1.0]), atol=1e-10)

    def test_p0_based_normalization_scale_by_magnitude(self):
        """Test p0-based normalization scales by initial parameter magnitudes."""
        # Setup: parameters with different magnitudes
        p0 = jnp.array([1000.0, 1.0, 0.001])

        # Create normalizer with p0 strategy (no bounds)
        normalizer = ParameterNormalizer(p0=p0, bounds=None, strategy="p0")

        # Normalize parameters
        normalized = normalizer.normalize(p0)

        # Expected: p0 / |p0| = [1.0, 1.0, 1.0] (normalized to magnitude 1)
        expected = jnp.array([1.0, 1.0, 1.0])
        np.testing.assert_allclose(normalized, expected, rtol=1e-7)

        # Check different parameters are scaled proportionally
        params_test = jnp.array([2000.0, 2.0, 0.002])
        normalized_test = normalizer.normalize(params_test)
        expected_test = jnp.array([2.0, 2.0, 2.0])
        np.testing.assert_allclose(normalized_test, expected_test, rtol=1e-7)

    def test_normalize_denormalize_are_inverses(self):
        """Test that normalize() and denormalize() are exact inverses."""
        # Test with bounds-based normalization
        p0 = jnp.array([5.0, 15.0, 25.0])
        bounds = (jnp.array([0.0, 10.0, 20.0]), jnp.array([10.0, 20.0, 30.0]))
        normalizer = ParameterNormalizer(p0=p0, bounds=bounds, strategy="bounds")

        # Round-trip: normalize then denormalize
        normalized = normalizer.normalize(p0)
        denormalized = normalizer.denormalize(normalized)
        np.testing.assert_allclose(denormalized, p0, rtol=1e-10)

        # Test with p0-based normalization
        p0_nobounds = jnp.array([100.0, 0.1, 10.0])
        normalizer_p0 = ParameterNormalizer(p0=p0_nobounds, bounds=None, strategy="p0")

        normalized_p0 = normalizer_p0.normalize(p0_nobounds)
        denormalized_p0 = normalizer_p0.denormalize(normalized_p0)
        np.testing.assert_allclose(denormalized_p0, p0_nobounds, rtol=1e-10)

    def test_normalization_jacobian_analytical(self):
        """Test normalization Jacobian is computed analytically and correct."""
        # Bounds-based: Jacobian should be diag(ub - lb)
        p0 = jnp.array([5.0, 15.0])
        bounds = (jnp.array([0.0, 10.0]), jnp.array([10.0, 20.0]))
        normalizer = ParameterNormalizer(p0=p0, bounds=bounds, strategy="bounds")

        jacobian = normalizer.normalization_jacobian

        # Expected: diagonal matrix with (ub - lb) on diagonal
        expected_diag = jnp.array([10.0, 10.0])
        expected_jacobian = jnp.diag(expected_diag)
        np.testing.assert_allclose(jacobian, expected_jacobian, rtol=1e-10)

        # p0-based: Jacobian should be diag(scale_factors)
        p0_nobounds = jnp.array([100.0, 0.01])
        normalizer_p0 = ParameterNormalizer(p0=p0_nobounds, bounds=None, strategy="p0")

        jacobian_p0 = normalizer_p0.normalization_jacobian

        # Expected: diagonal matrix with p0 magnitudes
        expected_diag_p0 = jnp.abs(p0_nobounds)
        expected_jacobian_p0 = jnp.diag(expected_diag_p0)
        np.testing.assert_allclose(jacobian_p0, expected_jacobian_p0, rtol=1e-10)

    def test_bounds_transformation_to_normalized_space(self):
        """Test bounds are correctly transformed to normalized space."""
        # Original bounds
        p0 = jnp.array([50.0, 0.5])
        bounds = (jnp.array([10.0, 0.0]), jnp.array([100.0, 1.0]))

        # Bounds-based normalization
        normalizer = ParameterNormalizer(p0=p0, bounds=bounds, strategy="bounds")

        lb_norm, ub_norm = normalizer.transform_bounds()

        # Bounds should map to [0, 1] for bounds-based
        np.testing.assert_allclose(lb_norm, jnp.array([0.0, 0.0]), atol=1e-10)
        np.testing.assert_allclose(ub_norm, jnp.array([1.0, 1.0]), atol=1e-10)

        # p0-based normalization
        p0_nobounds = jnp.array([10.0, 2.0])
        bounds_nobounds = (jnp.array([5.0, 1.0]), jnp.array([20.0, 4.0]))
        normalizer_p0 = ParameterNormalizer(
            p0=p0_nobounds, bounds=bounds_nobounds, strategy="p0"
        )

        lb_norm_p0, ub_norm_p0 = normalizer_p0.transform_bounds()

        # Bounds should be scaled by p0 magnitudes
        expected_lb = jnp.array([5.0 / 10.0, 1.0 / 2.0])  # [0.5, 0.5]
        expected_ub = jnp.array([20.0 / 10.0, 4.0 / 2.0])  # [2.0, 2.0]
        np.testing.assert_allclose(lb_norm_p0, expected_lb, rtol=1e-7)
        np.testing.assert_allclose(ub_norm_p0, expected_ub, rtol=1e-7)

    def test_normalization_disabled_identity_transform(self):
        """Test normalization='none' acts as identity transform."""
        # Create normalizer with 'none' strategy
        p0 = jnp.array([5.0, 15.0, 25.0])
        bounds = (jnp.array([0.0, 10.0, 20.0]), jnp.array([10.0, 20.0, 30.0]))
        normalizer = ParameterNormalizer(p0=p0, bounds=bounds, strategy="none")

        # Normalize and denormalize should return unchanged values
        normalized = normalizer.normalize(p0)
        np.testing.assert_allclose(normalized, p0, rtol=1e-10)

        denormalized = normalizer.denormalize(p0)
        np.testing.assert_allclose(denormalized, p0, rtol=1e-10)

        # Jacobian should be identity
        jacobian = normalizer.normalization_jacobian
        expected_jacobian = jnp.eye(len(p0))
        np.testing.assert_allclose(jacobian, expected_jacobian, rtol=1e-10)

    def test_auto_strategy_selection(self):
        """Test 'auto' strategy selects bounds or p0 appropriately."""
        # With bounds: should use bounds-based
        p0 = jnp.array([50.0, 0.5])
        bounds = (jnp.array([10.0, 0.0]), jnp.array([100.0, 1.0]))
        normalizer_with_bounds = ParameterNormalizer(
            p0=p0, bounds=bounds, strategy="auto"
        )

        # Should normalize to [0, 1] like bounds-based
        normalized = normalizer_with_bounds.normalize(p0)
        expected = jnp.array([40.0 / 90.0, 0.5])
        np.testing.assert_allclose(normalized, expected, rtol=1e-7)

        # Without bounds: should use p0-based
        p0_nobounds = jnp.array([100.0, 1.0])
        normalizer_no_bounds = ParameterNormalizer(
            p0=p0_nobounds, bounds=None, strategy="auto"
        )

        # Should normalize to magnitude 1 like p0-based
        normalized_nobounds = normalizer_no_bounds.normalize(p0_nobounds)
        expected_nobounds = jnp.array([1.0, 1.0])
        np.testing.assert_allclose(normalized_nobounds, expected_nobounds, rtol=1e-7)

    def test_zero_parameter_handling(self):
        """Test normalizer handles zero parameters gracefully."""
        # p0 with zero value
        p0 = jnp.array([10.0, 0.0, 5.0])

        # p0-based normalization should add epsilon to prevent division by zero
        normalizer = ParameterNormalizer(p0=p0, bounds=None, strategy="p0")

        # Should not raise error
        normalized = normalizer.normalize(p0)
        assert jnp.all(jnp.isfinite(normalized))

        # Zero parameter should be handled with small epsilon
        # The second element should be normalized to 0.0 (0 / epsilon ≈ 0)
        assert normalized[1] == 0.0


class TestNormalizedModelWrapper:
    """Tests for NormalizedModelWrapper class."""

    def test_wrapper_preserves_function_output(self):
        """Test wrapped model produces identical output to original."""

        # Define simple model
        def model(x, a, b, c):
            return a * jnp.exp(-b * x) + c

        # Create normalizer
        p0 = jnp.array([2.0, 0.5, 1.0])
        bounds = (jnp.array([0.0, 0.0, 0.0]), jnp.array([10.0, 2.0, 5.0]))
        normalizer = ParameterNormalizer(p0=p0, bounds=bounds, strategy="bounds")

        # Wrap model
        wrapped_model = NormalizedModelWrapper(model, normalizer)

        # Test data
        x = jnp.linspace(0, 5, 10)

        # Compute output with original model and original parameters
        original_output = model(x, 2.0, 0.5, 1.0)

        # Compute output with wrapped model and normalized parameters
        normalized_params = normalizer.normalize(p0)
        wrapped_output = wrapped_model(x, *normalized_params)

        # Outputs should be identical
        np.testing.assert_allclose(wrapped_output, original_output, rtol=1e-7)

    def test_jit_compatibility_of_wrapped_model(self):
        """Test wrapped model is JIT-compatible."""

        # Define model
        def model(x, a, b):
            return a * x + b

        # Create normalizer
        p0 = jnp.array([5.0, 10.0])
        normalizer = ParameterNormalizer(p0=p0, bounds=None, strategy="p0")

        # Wrap model
        wrapped_model = NormalizedModelWrapper(model, normalizer)

        # JIT compile wrapped model
        @jax.jit
        def jitted_wrapped_model(x, a_norm, b_norm):
            return wrapped_model(x, a_norm, b_norm)

        # Test data
        x = jnp.array([1.0, 2.0, 3.0])
        normalized_params = normalizer.normalize(p0)

        # Should compile and run without error
        output = jitted_wrapped_model(x, *normalized_params)

        # Verify output is correct
        expected = model(x, 5.0, 10.0)
        np.testing.assert_allclose(output, expected, rtol=1e-7)

        # Test that recompilation works with different shapes (static shapes)
        x2 = jnp.array([1.0, 2.0, 3.0, 4.0])
        output2 = jitted_wrapped_model(x2, *normalized_params)
        expected2 = model(x2, 5.0, 10.0)
        np.testing.assert_allclose(output2, expected2, rtol=1e-7)

    def test_wrapper_with_vectorized_operations(self):
        """Test wrapper works correctly with vmap and batch operations."""

        # Define model
        def model(x, a, b):
            return a * jnp.sin(b * x)

        # Create normalizer
        p0 = jnp.array([3.0, 2.0])
        normalizer = ParameterNormalizer(p0=p0, bounds=None, strategy="p0")

        # Wrap model
        wrapped_model = NormalizedModelWrapper(model, normalizer)

        # Test with vmap over x
        x_batch = jnp.array([[1.0], [2.0], [3.0]])
        normalized_params = normalizer.normalize(p0)

        # Vmap over first dimension of x
        vmapped_model = jax.vmap(
            lambda x_single: wrapped_model(x_single, *normalized_params)
        )

        output = vmapped_model(x_batch.squeeze())

        # Verify output
        expected = model(x_batch.squeeze(), 3.0, 2.0)
        np.testing.assert_allclose(output, expected, rtol=1e-7)

    def test_wrapper_gradient_computation(self):
        """Test gradients can be computed through wrapped model."""

        # Define model
        def model(x, a, b):
            return a * x**2 + b * x

        # Create normalizer
        p0 = jnp.array([2.0, 3.0])
        normalizer = ParameterNormalizer(p0=p0, bounds=None, strategy="p0")

        # Wrap model
        wrapped_model = NormalizedModelWrapper(model, normalizer)

        # Define loss function in normalized space
        def loss_fn(params_norm):
            x = jnp.array([1.0, 2.0, 3.0])
            y_true = jnp.array([5.0, 14.0, 27.0])  # a=2, b=3
            y_pred = wrapped_model(x, *params_norm)
            return jnp.sum((y_pred - y_true) ** 2)

        # Compute gradient
        normalized_params = normalizer.normalize(p0)
        grad = jax.grad(loss_fn)(normalized_params)

        # Gradient should be finite and non-zero (model is trainable)
        assert jnp.all(jnp.isfinite(grad))

        # At optimal parameters, gradient should be near zero
        # (though not exactly due to normalization)
        assert jnp.linalg.norm(grad) < 1e-10



"""Protocol contract tests for CacheProtocol.

This module tests that cache implementations conform to the
CacheProtocol and BoundedCacheProtocol defined in nlsq.interfaces.
"""

from typing import Any

import pytest

from nlsq.interfaces.cache_protocol import (
    BoundedCacheProtocol,
    CacheProtocol,
    DictCache,
)


class TestCacheProtocolDefinition:
    """Test that CacheProtocol is correctly defined."""

    def test_protocol_is_runtime_checkable(self):
        """CacheProtocol should be runtime_checkable."""

        class MockCache:
            def get(self, key: str, default: Any = None) -> Any:
                return default

            def set(self, key: str, value: Any) -> None:
                pass

            def clear(self) -> None:
                pass

        assert isinstance(MockCache(), CacheProtocol)

    def test_protocol_requires_all_methods(self):
        """Classes missing any method should not satisfy protocol."""

        class MissingGet:
            def set(self, key, value):
                pass

            def clear(self):
                pass

        class MissingSet:
            def get(self, key, default=None):
                return default

            def clear(self):
                pass

        class MissingClear:
            def get(self, key, default=None):
                return default

            def set(self, key, value):
                pass

        assert not isinstance(MissingGet(), CacheProtocol)
        assert not isinstance(MissingSet(), CacheProtocol)
        assert not isinstance(MissingClear(), CacheProtocol)


class TestDictCacheConformance:
    """Test that DictCache conforms to CacheProtocol."""

    def test_dict_cache_satisfies_protocol(self):
        """DictCache should satisfy CacheProtocol."""
        cache = DictCache()
        assert isinstance(cache, CacheProtocol)

    def test_dict_cache_get_default(self):
        """get() should return default for missing keys."""
        cache = DictCache()
        assert cache.get("missing") is None
        assert cache.get("missing", "default") == "default"

    def test_dict_cache_set_and_get(self):
        """set() and get() should work together."""
        cache = DictCache()
        cache.set("key", "value")
        assert cache.get("key") == "value"

    def test_dict_cache_clear(self):
        """clear() should remove all entries."""
        cache = DictCache()
        cache.set("a", 1)
        cache.set("b", 2)
        cache.clear()
        assert cache.get("a") is None
        assert cache.get("b") is None

    def test_dict_cache_contains(self):
        """__contains__ should work correctly."""
        cache = DictCache()
        assert "key" not in cache
        cache.set("key", "value")
        assert "key" in cache

    def test_dict_cache_len(self):
        """__len__ should return item count."""
        cache = DictCache()
        assert len(cache) == 0
        cache.set("a", 1)
        assert len(cache) == 1
        cache.set("b", 2)
        assert len(cache) == 2


class TestBoundedCacheProtocol:
    """Test BoundedCacheProtocol requirements."""

    def test_protocol_is_runtime_checkable(self):
        """BoundedCacheProtocol should be runtime_checkable."""

        class MockBoundedCache:
            def get(self, key: str, default: Any = None) -> Any:
                return default

            def set(self, key: str, value: Any) -> None:
                pass

            def clear(self) -> None:
                pass

            @property
            def size_bytes(self) -> int:
                return 0

            @property
            def max_size_bytes(self) -> int:
                return 1024

            def evict(self, n_bytes: int) -> int:
                return 0

        assert isinstance(MockBoundedCache(), BoundedCacheProtocol)

    def test_protocol_requires_size_properties(self):
        """Classes missing size properties should not satisfy protocol."""

        class MissingSizeBytes:
            def get(self, key, default=None):
                return default

            def set(self, key, value):
                pass

            def clear(self):
                pass

            @property
            def max_size_bytes(self):
                return 1024

            def evict(self, n_bytes):
                return 0

        assert not isinstance(MissingSizeBytes(), BoundedCacheProtocol)


class TestConcreteImplementations:
    """Test that concrete NLSQ cache implementations satisfy protocols."""

    def test_unified_cache_has_cache_methods(self):
        """UnifiedCache should have cache-like methods."""
        from nlsq.caching.unified_cache import UnifiedCache

        cache = UnifiedCache()
        # UnifiedCache is specialized for JIT caching
        # It has get_or_compile method for compiled function caching
        assert hasattr(cache, "get_or_compile")
        assert hasattr(cache, "clear")

    def test_smart_cache_exists(self):
        """SmartCache should exist and be importable."""
        from nlsq.caching.smart_cache import SmartCache

        cache = SmartCache()
        # SmartCache provides get/set methods
        assert hasattr(cache, "get")
        assert hasattr(cache, "set")


class TestCacheUsagePatterns:
    """Test common cache usage patterns."""

    def test_cache_as_lru(self):
        """Cache should work as simple LRU-like storage."""
        cache = DictCache()

        # Store multiple values
        for i in range(10):
            cache.set(f"key_{i}", f"value_{i}")

        # Retrieve values
        assert cache.get("key_5") == "value_5"
        assert cache.get("key_nonexistent") is None

    def test_cache_with_complex_values(self):
        """Cache should handle complex values."""
        import numpy as np

        cache = DictCache()

        # Store numpy array
        arr = np.array([1.0, 2.0, 3.0])
        cache.set("array", arr)
        retrieved = cache.get("array")
        np.testing.assert_array_equal(retrieved, arr)

        # Store dict
        cache.set("dict", {"a": 1, "b": [1, 2, 3]})
        assert cache.get("dict")["a"] == 1

    def test_cache_key_overwrite(self):
        """Setting same key should overwrite value."""
        cache = DictCache()

        cache.set("key", "value1")
        assert cache.get("key") == "value1"

        cache.set("key", "value2")
        assert cache.get("key") == "value2"



"""Focused tests for memory reuse optimizations (Task Group 5).

These tests validate the Phase 1 Priority 2 optimizations for memory reuse:
- Adaptive safety factor reduction (1.2 → 1.05)
- Size-class bucketing (1KB/10KB/100KB)
- Pool reuse statistics tracking
- disable_padding configuration flag
"""

import unittest

import jax.numpy as jnp

from nlsq.caching.memory_manager import MemoryManager, get_memory_manager
from nlsq.caching.memory_pool import MemoryPool, clear_global_pool, get_global_pool


class TestAdaptiveSafetyFactor(unittest.TestCase):
    """Test adaptive safety factor reduction from 1.2 → 1.05."""

    def setUp(self):
        """Set up test fixtures."""
        self.manager = MemoryManager(safety_factor=1.2)

    def test_initial_safety_factor(self):
        """Test initial safety factor is 1.2."""
        self.assertEqual(self.manager.safety_factor, 1.2)

    def test_adaptive_reduction_after_warmup(self):
        """Test safety factor reduces to ~1.05 after successful runs."""
        # Create manager with adaptive safety enabled
        manager = MemoryManager(safety_factor=1.2, enable_adaptive_safety=True)

        # Simulate successful memory allocations
        for _ in range(15):
            with manager.memory_guard(bytes_needed=1000000):
                # Simulate using less memory than predicted (typical case)
                pass

        # After warmup (10 runs), safety factor should start reducing
        telemetry = manager.get_safety_telemetry()
        self.assertGreater(telemetry["telemetry_entries"], 10)

        # Safety factor should have reduced from initial 1.2
        self.assertLess(manager.safety_factor, 1.2)
        # But not below minimum of 1.05
        self.assertGreaterEqual(manager.safety_factor, 1.05)

    def test_safety_factor_in_memory_prediction(self):
        """Test safety factor is applied in memory predictions."""
        # Predict memory with safety factor
        memory_with_safety = self.manager.predict_memory_requirement(
            n_points=1000, n_params=5, algorithm="trf", dtype=jnp.float64
        )

        # Create manager with safety_factor=1.0 for baseline
        manager_no_safety = MemoryManager(safety_factor=1.0)
        memory_no_safety = manager_no_safety.predict_memory_requirement(
            n_points=1000, n_params=5, algorithm="trf", dtype=jnp.float64
        )

        # With safety factor should be larger
        self.assertGreater(memory_with_safety, memory_no_safety)
        # Should be approximately 1.2x larger
        ratio = memory_with_safety / memory_no_safety
        self.assertAlmostEqual(ratio, 1.2, delta=0.01)


class TestSizeClassBucketing(unittest.TestCase):
    """Test size-class bucketing for memory pool (1KB/10KB/100KB)."""

    def setUp(self):
        """Set up test fixtures."""
        self.pool = MemoryPool(max_pool_size=50, enable_stats=True)

    def tearDown(self):
        """Clean up after tests."""
        self.pool.clear()

    def test_bucket_1kb_small_arrays(self):
        """Test 1KB bucketing for small arrays."""
        # Small arrays should bucket to nearest 1KB
        # This will be implemented when bucketing is added (Task 5.4)
        # For now, verify basic allocation works
        arr = self.pool.allocate((100,), dtype=jnp.float64)  # 800 bytes
        self.assertEqual(arr.shape, (100,))

    def test_bucket_10kb_medium_arrays(self):
        """Test 10KB bucketing for medium arrays."""
        # Medium arrays should bucket to nearest 10KB
        arr = self.pool.allocate((1000,), dtype=jnp.float64)  # 8KB
        self.assertEqual(arr.shape, (1000,))

    def test_bucket_100kb_large_arrays(self):
        """Test 100KB bucketing for large arrays."""
        # Large arrays should bucket to nearest 100KB
        arr = self.pool.allocate((10000,), dtype=jnp.float64)  # 80KB
        self.assertEqual(arr.shape, (10000,))

    def test_bucketing_increases_reuse_rate(self):
        """Test that bucketing increases pool reuse rate by 5x."""
        # Allocate and release arrays with slightly different sizes
        sizes = [95, 100, 105, 98, 102]  # All should bucket to same size

        for size in sizes:
            arr = self.pool.allocate((size,), dtype=jnp.float64)
            self.pool.release(arr)

        # After bucketing implementation, these should all reuse the same buffer
        # For now, verify releases work
        stats = self.pool.get_stats()
        self.assertEqual(stats["releases"], len(sizes))


class TestPoolReuseStatistics(unittest.TestCase):
    """Test pool reuse statistics tracking."""

    def setUp(self):
        """Set up test fixtures."""
        self.pool = MemoryPool(max_pool_size=10, enable_stats=True)

    def tearDown(self):
        """Clean up after tests."""
        self.pool.clear()

    def test_reuse_rate_calculation(self):
        """Test reuse_rate = reused_allocations / total_allocations."""
        # First allocation (no reuse)
        arr1 = self.pool.allocate((10,), dtype=jnp.float64)
        self.pool.release(arr1)

        # Second allocation (should reuse)
        arr2 = self.pool.allocate((10,), dtype=jnp.float64)
        self.pool.release(arr2)

        stats = self.pool.get_stats()
        total = stats["allocations"] + stats["reuses"]
        expected_rate = stats["reuses"] / total if total > 0 else 0.0

        self.assertEqual(stats["reuse_rate"], expected_rate)
        self.assertGreater(stats["reuse_rate"], 0)

    def test_pool_stats_in_diagnostics(self):
        """Test pool statistics are included in optimization diagnostics."""
        # Allocate and reuse multiple times
        for _ in range(5):
            arr = self.pool.allocate((20,), dtype=jnp.float64)
            self.pool.release(arr)

        stats = self.pool.get_stats()

        # Verify all expected statistics are present
        self.assertIn("reuse_rate", stats)
        self.assertIn("allocations", stats)
        self.assertIn("reuses", stats)
        self.assertIn("releases", stats)
        self.assertIn("currently_allocated", stats)

    def test_peak_pool_size_tracking(self):
        """Test peak pool size is tracked correctly."""
        # Allocate multiple different shapes
        arrs = [
            self.pool.allocate((10,), dtype=jnp.float64),
            self.pool.allocate((20,), dtype=jnp.float64),
            self.pool.allocate((30,), dtype=jnp.float32),
        ]

        stats = self.pool.get_stats()
        self.assertGreater(stats["currently_allocated"], 0)
        self.assertGreaterEqual(stats["currently_allocated"], len(arrs))


class TestDisablePaddingFlag(unittest.TestCase):
    """Test disable_padding configuration flag."""

    def test_padding_enabled_by_default(self):
        """Test padding is enabled by default."""
        manager = MemoryManager()
        # Default behavior: safety_factor > 1.0 implies padding enabled
        self.assertGreater(manager.safety_factor, 1.0)

    def test_disable_padding_sets_exact_shapes(self):
        """Test disable_padding uses exact shapes (no bucketing)."""
        # This will be implemented when disable_padding is added (Task 5.6)
        # For now, verify manager can be created with different safety factors
        manager_padded = MemoryManager(safety_factor=1.2)
        manager_exact = MemoryManager(safety_factor=1.0)

        memory_padded = manager_padded.predict_memory_requirement(
            n_points=1000, n_params=5, algorithm="trf"
        )
        memory_exact = manager_exact.predict_memory_requirement(
            n_points=1000, n_params=5, algorithm="trf"
        )

        # Exact should be less than padded
        self.assertLess(memory_exact, memory_padded)

    def test_disable_padding_for_strict_environments(self):
        """Test disable_padding works correctly for cloud quotas."""
        # When disable_padding=True: exact shapes, safety_factor=1.0
        manager = MemoryManager(safety_factor=1.0)

        memory = manager.predict_memory_requirement(
            n_points=1000, n_params=5, algorithm="trf"
        )

        # Verify it's not over-allocating
        self.assertIsInstance(memory, int)
        self.assertGreater(memory, 0)


class TestMemoryUsageReduction(unittest.TestCase):
    """Test 10-20% peak memory reduction on standard fits."""

    def setUp(self):
        """Set up test fixtures."""
        self.manager_original = MemoryManager(safety_factor=1.2)
        self.manager_optimized = MemoryManager(safety_factor=1.05)

    def test_memory_reduction_with_adaptive_safety(self):
        """Test memory usage reduces by 10-20% with adaptive safety factor."""
        # Standard fit scenario
        n_points = 10000
        n_params = 10

        memory_original = self.manager_original.predict_memory_requirement(
            n_points=n_points, n_params=n_params, algorithm="trf"
        )

        memory_optimized = self.manager_optimized.predict_memory_requirement(
            n_points=n_points, n_params=n_params, algorithm="trf"
        )

        reduction = (memory_original - memory_optimized) / memory_original
        # With safety factor 1.2 → 1.05, expect ~12.5% reduction
        self.assertGreater(reduction, 0.10)  # At least 10%
        self.assertLess(reduction, 0.20)  # No more than 20%

    def test_mixed_precision_memory_savings(self):
        """Test mixed precision provides additional 50% memory savings."""
        # float64 baseline
        memory_f64 = self.manager_original.predict_memory_requirement(
            n_points=10000, n_params=10, algorithm="trf", dtype=jnp.float64
        )

        # float32 should be ~50% smaller
        memory_f32 = self.manager_original.predict_memory_requirement(
            n_points=10000, n_params=10, algorithm="trf", dtype=jnp.float32
        )

        reduction = (memory_f64 - memory_f32) / memory_f64
        self.assertAlmostEqual(reduction, 0.5, delta=0.05)  # ~50% reduction


class TestMemoryReuseIntegration(unittest.TestCase):
    """Integration tests for memory reuse across components."""

    def setUp(self):
        """Clear global pool before each test."""
        clear_global_pool()

    def test_global_memory_manager_integration(self):
        """Test global memory manager can be used across optimization runs."""
        manager = get_memory_manager()
        self.assertIsInstance(manager, MemoryManager)

        # Should return same instance
        manager2 = get_memory_manager()
        self.assertIs(manager, manager2)

    def test_pool_and_manager_coordination(self):
        """Test memory pool and manager coordinate correctly."""
        pool = get_global_pool(enable_stats=True)
        manager = get_memory_manager()

        # Allocate, release, then allocate again to create reuse
        arr1 = pool.allocate((100,), dtype=jnp.float64)
        pool.release(arr1)
        arr2 = pool.allocate((100,), dtype=jnp.float64)
        pool.release(arr2)

        stats_pool = pool.get_stats()
        stats_manager = manager.get_memory_stats()

        self.assertGreater(stats_pool["reuse_rate"], 0)
        self.assertIsInstance(stats_manager, dict)


if __name__ == "__main__":
    unittest.main()
